import json
import os.path
import re
import uuid
from datetime import datetime
from typing import Optional

import dateutil.parser
from jsonschema.validators import validate

from chemotion_api.elements.analyses_manager import AnalysesManager

from chemotion_api.connection import Connection
from chemotion_api.elements.attachments import Attachments, MutableAttachments

from chemotion_api.generic_segments import GenericSegments
from chemotion_api.utils import  parse_generic_object_json, \
    clean_generic_object_json, merge_dicts, snake_to_camel_case, fixeddict_serializer, PropertyDict

from requests.exceptions import RequestException


class Dataset(dict):
    """
    A Dataset is an object that consists of a metadata set and a measurement file.
    If all components have been configured correctly in Chemotion, a BagIT
    container is created when a new file is uploaded. This container is then used
    to automatically fill in the metadata set and forward the measurement data to ChemSpectra.

    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> s = instance.get_sample(1)
    >>> # Loads dataset V2024_08_23 from 'Analyses MT 1'
    >>> d1 = s.analyses.by_name('Analyses MT 1').dataset_by_name('V2024_08_23')
    >>> # Write BagIT zip and metadata set as xlsx file
    >>> d1.write_data_set_xlsx('./CWD')
    >>> d1.write_zip('./CWD')
    """

    def __init__(self, session: Connection, json_data: dict, attachments: MutableAttachments = None):
        self.id = json_data.get('id')
        self.name = json_data.get('name')
        self.description = json_data.get('description')
        ds_json = json_data.get('dataset')
        if ds_json is not None:
            res = parse_generic_object_json(ds_json)
            super().__init__(res.get('values'))
            self._mapping = res.get('obj_mapping')
        self._session = session
        self._json_data = json_data
        self._attachments = attachments

    @property
    def json_data(self) -> dict:
        return self._json_data

    @property
    def attachments(self) -> MutableAttachments:
        """
        A list of all attached files. It contains the BagIt, the ChemSpectra results
         as well as the measurement file

        :return: A mutable attachment File
        """
        if self._attachments is None:
            attachments_data = self._json_data.get('attachments')
            if attachments_data is not None:
                self._attachments = MutableAttachments(self._session, attachments_data)
        return self._attachments

    def write_zip(self, destination=''):
        """
        Writes the BagIT zip file.

        :param destination: Director where the BagIT should be saved to
        :return: Abs filepath of the BagIT zip
        """
        image_url = "/api/v1/attachments/zip/{}".format(self.id)
        res = self._session.get(image_url)
        if res.status_code != 200:
            raise ConnectionRefusedError('{} -> {}'.format(res.status_code, res.text))

        if not os.path.exists(destination) or os.path.isdir(destination):
            regex_file_name = re.search('filename="([^"]+)', res.headers['Content-Disposition'])
            destination = os.path.join(destination, regex_file_name.groups()[0])

        with open(destination, 'wb+') as f:
            f.write(res.content)

        return destination

    def write_data_set_xlsx(self, destination=''):
        """
        Writes the metadata set as xlsx.

        :param destination: Director where the BagIT should be saved to.
        :return: Abs filepath of the BagIT zip.
        """
        image_url = "/api/v1/export_ds/dataset/{}".format(self.id)
        res = self._session.get(image_url)
        if res.status_code != 200:
            raise ConnectionRefusedError('{} -> {}'.format(res.status_code, res.text))

        if not os.path.exists(destination) or os.path.isdir(destination):
            regex_file_name = re.search('filename="([^"]+)', res.headers['Content-Disposition'])
            destination = os.path.join(destination, regex_file_name.groups()[0])

        with open(destination, 'wb+') as f:
            f.write(res.content)

        return destination

    def to_clean_json(self):
        """
        Cleans the data as perpetration to be saved
        """
        ds = self._json_data.get('dataset')
        if ds is not None:
            clean_generic_object_json(ds, self, self._mapping)
            ds['changed'] = True

    def save_attachments(self):
        self.attachments.save()
        self.json_data['attachments'] = [x for x in self.attachments]


class Analyses(dict):
    """
    An analysis consists of a list of metadata schemas and associated data files.
    The schemas can be edited. And the files can be read and saved to your local device
    """

    def __init__(self, data: dict, session: Connection):
        super().__init__()
        self._session = session
        self.id = data.get('id')
        self.type = data.get('extended_metadata', {}).get('kind', '')

        self._data = data
        self['name'] = data['name']
        self['description'] = data['description']

        for k in ['report', 'status', 'kind', 'content']:
            if k in data['extended_metadata']:
                self[k] = data['extended_metadata'][k]
        self.datasets = []
        for jd in self._data.get('children'):
            self.datasets.append(Dataset(session, jd))

    @property
    def is_new(self) -> bool:
        """
        :return: Returns true if the analyses has not been saved yet.
        """
        return self._data.get('is_new', False)

    def add_dataset(self, file_path: str) -> Dataset:
        """
        Loads a datafile up  and creates a new dataset.
        The Chemotion Instance will convert the date file and create a metadate set accordingly.
        However, this process only works if a Converter profile has been created

        :param file_path: the path on your local device to the datafile
        :return: A new Dataset object
        """
        ds = AnalysesManager.build_new_dataset()
        ds['name'] = os.path.basename(file_path)
        d = Dataset(self._session, ds, MutableAttachments(self._session, ds['attachments'], file_path))
        self.datasets.append(d)
        self._data['children'].append(d.json_data)
        return d

    def dataset_by_name(self, name: str) -> Dataset:
        """
        Find a dataset by its name.

        :param name: Name of the dataset
        :return: A Dataset object
        """
        return next(x for x in self.datasets if x.name == name)

    def preview_image(self) -> bytes | None:
        """
        Loads the image used as preview for the analyses.

        Usage::

        >>> from chemotion_api import Instance
        >>> import logging
        >>> try:
        >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
        >>> except ConnectionError as e:
        >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
        >>> s = instance.get_sample(1)
        >>> # Loads analysis 'Analyses MT 1'
        >>> analyses = s.analyses.by_name('Analyses MT 1')
        >>> with open('./img.svg', 'wb+') as f:
        >>>     f.write(analyses.preview_image())
        >>> s.save()

        :raises ConnectionError: If the request fails
        :return: byte representation of an image
        """

        if self._data.get('preview_img') is None or self._data.get('preview_img').get('id') is None:
            return None
        return self._load_image(self._data.get('preview_img').get('id'))

    def _load_image(self, file_id: int) -> bytes:
        image_url = "/api/v1/attachments/{}".format(file_id)
        res = self._session.get(image_url)
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        return res.content

    def to_json(self):
        """
        A savable representation of the Analyses
        :return: Json representation
        """
        self._data['name'] = self['name']
        self._data['description'] = self['description']
        for k in ['report', 'status', 'kind', 'content']:
            if k in self:
                self._data['extended_metadata'][k] = self[k]
        for ds in self.datasets:
            ds.to_clean_json()
        return self._data

    def save_attachments(self):
        """
        Do not use this methode! It is automatically used before the element is saved!
        """

        for i, d in enumerate(self.datasets):
            d.save_attachments()

    def clean_up_after_save(self):
        """
        Do not use this methode! It is automatically used after the element is saved!
        """

        self._data['is_new'] = False
        for d in self.datasets:
            d.json_data['is_new'] = False


class AnalysesList(list):
    """
    List of :class:`chemotion_api.elements.abstract_element.Analyses`.


    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> s = instance.get_sample(1)
    >>> # Download of all analysis files of data set 'V2024_08_23' from 'Analyses MT 1'
    >>> analyses = s.analyses.by_name('Analyses MT 1')
    >>> dataset = analyses.dataset_by_name('V2024_08_23')
    >>> os.makedirs('./TO_DEL', exist_ok=True)
    >>> for atta in dataset.attachments:
    >>>     atta.save_file('./dataset_files')
    >>> # Change value in the metadate set
    >>> dataset['General Information']['Label'] = 'V2024_08_23'
    >>> # Create a new
    >>> ana = s.analyses.add_analyses("Analyses MT 2")
    >>> ana.add_dataset('./V2024_08_24.json')
    >>> s.save()
    """

    def __init__(self, conn: Connection):
        super().__init__()
        self.conn = conn

    def by_name(self, name: str) -> Analyses:
        return next(x for x in self if x['name'] == name)

    def by_id(self, id: str | int) -> Analyses:
        return next(x for x in self if str(x.id) == str(id))

    def add_analyses(self, name: str = 'New') -> Analyses:
        new_a = Analyses(AnalysesManager.build_new_analyses(), self.conn)
        self.append(new_a)
        new_a['name'] = name
        return new_a

    def save_attachments(self):
        for ana in self:
            ana.save_attachments()

    def clean_up_after_save(self):
        for ana in self:
            ana.clean_up_after_save()


class Segment:
    def __init__(self, generic_segment: dict, segment_data: Optional[dict]):
        self._segment_data = segment_data
        self._generic_segment = generic_segment
        self._properties = None
        self._properties_mapping = None
        self._validator = None
        if self._segment_data is not None:
            self._prepare_data()

    @property
    def layer(self):
        if self._properties is None:
            return None
        return PropertyDict(self.on_properties_change, self._properties)

    def _prepare_data(self):
        temp_segment = parse_generic_object_json(self._segment_data)
        self._properties = temp_segment.get('values')
        self._properties_mapping = temp_segment.get('obj_mapping')
        self._validator = temp_segment.get('validator')

    def load(self):
        self._segment_data = GenericSegments.new_session(self._generic_segment)
        self._prepare_data()

    def on_properties_change(self, dict_obj, name, value):
        dict_obj[-1][name[-1]] = self._validator.validate(name[0], name[1], value[-1])
        return False

    @property
    def is_loaded(self):
        return self._segment_data is not None

    def clean(self):
        clean_generic_object_json(self._segment_data, self._properties, self._properties_mapping)
        return self._segment_data


class Segments(PropertyDict):
    def __init__(self, generic_segments: GenericSegments, element_type: str, segments_json: list[dict]):
        super().__init__(None, {})
        self._generic_segments = generic_segments
        self._all_seg_classes = [seg for seg in generic_segments.all_classes if
                                 seg.get('element_klass').get('name') == element_type]
        self._element_type = element_type
        self._populate_segments(segments_json)


    def _populate_segments(self, segments_json):
        for seg in self._all_seg_classes:
            self.data[seg.get('label')] = Segment(seg,
                                                  next((x for x in segments_json if x.get('segment_klass_id') == seg.get('id')),
                                                       None))

    def get(self, key):
        val: Segment = super().get(key)
        if not val.is_loaded:
            val.load()
        return val

    def clean(self):
        return_val = []
        for seg_label, seg in self.data.items():
            if seg.is_loaded:
                return_val.append(seg.clean())
        return return_val


class AbstractElement:
    """
    This abstract element is the basis for all chemotion-api elements. It provides all the necessary functions and properties to work with these elements.
    It contains. For more details check the Element classes out.
    """

    def __init__(self, generic_segments: GenericSegments, session: Connection, json_data: dict = None, id: int = None,
                 element_type: str = None, avoid_load: bool = False):
        self._attachments: Attachments | None = None
        self._generic_segments = generic_segments
        self._session = session

        self._short_label = None
        self._properties = None
        self._analyses = None
        self._segments = None
        # Name of element instance
        self._name: str = ''
        # Last update of the element
        self.last_update: datetime | None = None
        self._element_type = element_type
        self.json_ld: dict = {}
        # Json data contains the raw data fetched from the server
        self.json_data: dict | None = None
        self.id = id or uuid.uuid4().__str__()
        # Json LD description of the element. It contains @type and @ID

        self._data_loaded = False
        self._properties = None
        if avoid_load:
            self._properties = {}
            return
        if json_data is not None:
            self._set_json_data(json_data)
        elif not self.is_new and element_type is not None:
            self.load()
        else:
            raise ValueError("Either 'json_data' or 'id' and 'element_type' must be provided during initialization")

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name
        if self.json_data and 'name' in self.json_data:
            self.json_data['name'] = name

    @property
    def short_label(self) -> str | None:
        """
        The short label of en element if available

        :return: short label of en element
        """
        return self._short_label

    @property
    def properties(self) -> PropertyDict:
        """
        The properties property contains all data from the main tab of the elements in Chemotion.

        :return: Element properties
        """
        return PropertyDict(self.on_properties_change, self._properties)

    @property
    def analyses(self) -> AnalysesList:
        """
        With the analyses object one can read and write all information from the analysis Tab in Chemotion.

        :return: Analyses object
        """
        return self._analyses

    @property
    def segments(self) -> Segments:
        """
        Contains all segments (tabs) of elements in Chemotion.
        It contains generic elements as well as the proerties and the analyses segment

        :return: Segment object
        """
        return self._segments

    @property
    def attachments(self) -> Attachments | None:
        """
        Attachment container of the element. This s None if the element has no Attachments

        :return: Attachments container
        """
        if self._attachments is None:
            attachments_data = self.json_data.get('attachments')
            if attachments_data is not None:
                self._attachments = Attachments(self._session, attachments_data)
        return self._attachments

    @property
    def id(self) -> int:
        """
        Database ID of the element

        :return: Database ID of the element
        """
        return self._id

    @id.setter
    def id(self, id_val: Optional[int | str]):
        """
        Database ID of the element
        """
        if id_val is None:
            self.is_new = True
        elif isinstance(id_val, str):
            if id_val.isdigit():
                id_val = int(id_val)
                self.is_new = False
            else:
                self.is_new = True
        elif isinstance(id_val, int):
            self.is_new = False

        self._id = id_val
        if self.json_data is not None:
            self.json_ld[
                '@id'] = f"{self._session._host_url}{self.get_url(self.element_type)}/{self.id}.json" if self.id is not None else uuid.uuid4().__str__()

    @property
    def is_generic(self):
        return self.__class__.__name__ == 'GenericElement'

    @property
    def element_type(self) -> str:
        """
        Element type: is either 'sample', 'sample', 'reaction', 'wellplate', 'research_plan' or if the element is a generic one it is only 'element'

        :return: The element type
        """
        return self._element_type

    @element_type.setter
    def element_type(self, element_type_val: str):
        """
        Element type: is either 'sample', 'sample', 'reaction', 'wellplate', 'research_plan' or if the element is a generic one it is only 'element'
        Has to be set if json data is not available at the moment of creation
        """
        self._element_type = element_type_val
        if self.json_data is not None:
            schema_version = self.json_data.get('properties', {}).get('version')
            self.json_ld['@type'] = self._session.schema_manager.generate_model_type(self.element_type, schema_version)

    def copy_params(self) -> dict:
        return {'generic_segments': self._generic_segments,
                'element_type': self._element_type, 'avoid_load': False,
                'session': self._session, 'json_data': self.json_data, 'id': self.id}

    def __copp__(self):
        return self.__class__(**self.copy_params())

    def properties_schema(self) -> dict:
        """
        Returns the JSON.org schema of the cleaned properties.

        :return: JSON.org schema
        """
        return {}

    def load(self):
        """
        Loads and parses the data of an element from the server.
        The standard Values can be found in the properties property or the segment

        :raises RequestException: If the element could not be loaded!

        :raises ValueError: If no ID available!
        """

        self._data_loaded = True

        attachments_data, json_data = self._load_data_from_server()
        if attachments_data is not None:
            self._attachments = Attachments(self._session, attachments_data)
        self._set_json_data(json_data)

    def _load_data_from_server(self):

        if self.is_new:
            raise ValueError("No ID available")
        payload = {}
        res = self._session.get(f"{self.__class__.get_url(self.element_type)}/{self.id}.json",
                                data=payload)
        if res.status_code != 200:
            raise RequestException("{} -> {}".format(res.status_code, res.text))
        res_json = res.json()
        attachments_data = res_json.get('attachments')
        return attachments_data, res_json[self.get_response_key(self.element_type)]

    def _set_json_data(self, json_data):
        self.json_data = json_data
        if json_data.get('updated_at') is not None:
            try:
                for pattern in ['%Y-%m-%d %H:%M:%S %Z', '%d.%m.%Y, %H:%M:%S %z']:
                    try:
                        self.last_update = datetime.strptime(json_data.get('updated_at'), pattern)
                        break
                    except:
                        pass

                if self.last_update is None:
                    self.last_update = dateutil.parser.parse(json_data.get('updated_at'))
            except:
                pass
        self.name = json_data.get('name', '')

        self._short_label = self.json_data.get('short_label')
        self.element_type = json_data.get('type')
        self.id = json_data.get('id', self._id)

        self._properties: dict = self._parse_properties()
        self._analyses: AnalysesList = self._parse_analyses()
        segment_data = self.json_data.get('segments', [])

        self._segments = Segments(self._generic_segments,
                                  json_data.get('type'),
                                  segment_data)

    def save(self, load_after_save: bool = False):
        """
        Saves or creates an object according to the set properties. It overwrites the
        json_data entries by the values set in the segments object.

        :raises RequestException: If request was not successful
        :raises jsonschema.exceptions.ValidationError: If the properties ar not valid
        """
        if not self._data_loaded and not self.is_new:
            _attachment_data, json_data = self._load_data_from_server()
            self.json_data = json_data

        self._data_loaded = True
        self.analyses.save_attachments()
        data = self.clean_data()
        is_created = False
        json_dumps_data = json.dumps(data, default=fixeddict_serializer)
        # Uncomment  after Chem update to write upload json to compare with Chemotion Browser upload
        # dump_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'test_config/dumps/upload_jsons')
        # os.makedirs(dump_dir, exist_ok=True)
        # with open(os.path.join(dump_dir, f'{self.element_type}_{self.id}_{uuid.uuid4().__str__()}.json'), 'w+') as outfile:
        #     outfile.write(json_dumps_data)
        if self.is_new:
            res = self._session.post(self.save_url(), data=json_dumps_data)
            is_created = True
        else:
            res = self._session.put(self.save_url(), data=json_dumps_data)
        if res.status_code != 200 and res.status_code != 201:
            raise RequestException(f'{res.status_code} -> {res.text}')
        if is_created:
            res_json = res.json()
            res_json = merge_dicts(self.json_data, res_json.get(self.element_type, res_json.get('element')))
            self._set_json_data(res_json)
        if self.attachments is not None:
            self.attachments.save(self.id, snake_to_camel_case(self.get_response_key(self.element_type)))
        self.analyses.clean_up_after_save()
        if load_after_save:
            self.load()

    def clean_data(self) -> dict:
        """
        Takes the values from the segments object and attachment object
        and overwrites the json_data values accordingly.

        :raises jsonschema.exceptions.ValidationError: If the properties ar not valid

        :return: cleaned data
        """
        cleaned_data = {
            'id': self.id or uuid.uuid4().__str__(),
            'is_new': self.is_new,
            'type': self.element_type
        }
        if isinstance(self.name, str):
            cleaned_data['name'] = self.name
        if 'collection_id' in self.json_data:
            cleaned_data['collection_id'] = self.json_data['collection_id']
        if 'container' in self.json_data:
            cleaned_data['container'] = self.json_data['container']
        if self.attachments is not None:
            cleaned_data['attachments'] = self.attachments.attachment_data

        cleaned_prop_data = self._clean_properties_data()
        for key, value in cleaned_prop_data.items():
            if isinstance(value, PropertyDict):
                cleaned_prop_data[key] = value.to_dict()
        validate(cleaned_prop_data, self.properties_schema())

        merge_dicts(cleaned_data, self._clean_segments_data(), cleaned_prop_data,
                    self._clean_analyses_data())
        return cleaned_data

    def save_url(self) -> str:
        """
        Retrieves the save URL, which varies based on whether the element already has an ID.

        :return: The save URL
        """
        if not self.is_new:
            return "/api/v1/{}s/{}".format(self.json_data.get('type'), self.id)
        return "/api/v1/{}s/".format(self.json_data.get('type'))

    def on_properties_change(self, dict_obj, name, value):
        return True

    def __eq__(self, other):
        return isinstance(other, AbstractElement) and self.json_ld['@id'] == other.json_ld['@id']

    def __ne__(self, other):
        return not isinstance(other, AbstractElement) or self.json_ld['@id'] != other.json_ld['@id']

    def _parse_properties(self) -> dict:
        raise NotImplemented

    def _clean_properties_data(self, serialize_data: dict | None = None) -> dict:
        raise NotImplemented

    def _parse_analyses(self) -> AnalysesList:
        analyses_list = AnalysesList(self._session)
        container = self.json_data.get('container')
        if container is not None and len(container.get('children', [])) > 0:
            for analyses in container.get('children', [{}])[0].get('children', []):
                analyses_list.append(Analyses(analyses, self._session))
        return analyses_list

    def _clean_analyses_data(self) -> dict:
        container = self.json_data.get('container')
        if container is None:
            return {}
        obj = {"container": {
            "children": [{"children": []}]
        }}
        res_list = container.get('children', [{}])[0].get('children', [])
        for (idx, analyses) in enumerate(res_list):
            analyses_obj: list[Analyses] = [item for (index, item) in enumerate(self.analyses) if
                                            item.id == analyses.get('id')]
            if len(analyses_obj) == 1:
                new_data = analyses_obj[0].to_json()
                for (key, item) in analyses.items():
                    if key in new_data:
                        res_list[idx][key] = new_data.get(key, res_list[idx][key])
        res_list += [item.to_json() for item in self.analyses if item.is_new]
        obj["container"]["children"][0]["children"] = res_list
        return obj

    def _clean_segments_data(self) -> dict:
        return {'segments': self.segments.clean()}

    @classmethod
    def get_response_key(cls, name: str) -> str:
        """
        Returns the element name used to construct the save and load URL

        :param name: The name of the element type

        :return: 
        """
        if name == 'sample':
            return 'sample'
        elif name == 'reaction':
            return 'reaction'
        elif name == 'wellplate':
            return 'wellplate'
        elif name == 'research_plan':
            return 'research_plan'
        return 'element'

    @classmethod
    def get_url(cls, name: str) -> str:
        """
        Returns the load URL based on the element type

        :param name: The name of the element type

        :return: The load url
        """
        if name == 'sample':
            return '/api/v1/samples'
        elif name == 'reaction':
            return '/api/v1/reactions'
        elif name == 'wellplate':
            return '/api/v1/wellplates'
        elif name == 'research_plan':
            return '/api/v1/research_plans'
        return f'/api/v1/generic_elements'



import copy

from chemotion_api.labImotion.items.validation.registry import SchemaRegistry
from chemotion_api.labImotion.items.validation.schemas.abstract_element_schema import generic_element_schema

generic_schema = copy.deepcopy(generic_element_schema)
generic_schema['$id'] = "chemotion://generic/element/draft-01"
generic_schema['title'] = "Schema for generic element"
generic_schema['required'] += ['name', 'icon_name', 'is_generic', 'klass_prefix']

generic_schema['properties'] |= {
    "name": {
        "type": "string"
    },
    "icon_name": {
        "type": "string"
    },
    "klass_prefix": {
        "type": "string"
    },
    "is_generic": {
        "type": "boolean"
    }
}

SchemaRegistry.instance().register(generic_schema)



from referencing import Registry, Resource
from jsonschema.validators import validator_for


class SchemaRegistry:
    """
    This calls manages all reader. It must be used as Singleton
    """
    _instance = None
    _schema_store = []
    _registry = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls):
        """
        Singleton constructor
        :return: Readers Singleton
        """
        if cls._instance is None:
            cls._instance = cls.__new__(cls)

        return cls._instance

    @property
    def registry(self):
        return self._schema_store

    def validator_for(self, id: str):
        if self._registry is None:
            self._registry = Registry().with_resources(self._schema_store)
        schema = self._registry.get(id).contents
        return validator_for(schema)(schema, registry=self._registry)


    def register(self, schema: dict):
        """
        Register a shcema
        :param reader:
        :param json_schema: schema.org schema
        :return:
        """
        self._schema_store.append((schema["$id"], Resource.from_contents(schema)))



import uuid
import re
from enum import Enum

from chemotion_api.labImotion.items.fields.abstract_fields import SystemDefined
from chemotion_api.labImotion.items.options import SubFieldType, FieldUnit, get_unit
from chemotion_api.labImotion.structure import StructureRegistry
from chemotion_api.labImotion.utils import ChemotionSerializable


class SubBaseField(ChemotionSerializable):
    _default_required_fields = ["id", "field_type", "value"]
    required_fields = []
    mapping = {"field_type": "type"}
    default = ""
    field_type = None

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        self._sr = sr
        self._layer_key = layer_key
        self._id = json_data.get("id", uuid.uuid4().__str__())
        self.value = json_data.get("value", self.default)

    def to_dict(self):
        json_dict = {}
        for key in self._default_required_fields + self.required_fields:
            val = getattr(self, key)
            if isinstance(val, Enum):
                val = val.value
            key = self.mapping.get(key, key)
            json_dict[key] = val
        return json_dict

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"value {type(value)} not supported")
        self._value = value

    @property
    def id(self) -> str:
        return self._id

    @id.setter
    def id(self, value: str):
        if not isinstance(id, str):
            raise TypeError(f"value {type(id)} not supported")
        self._id = id


class NumberField(SubBaseField):
    field_type = SubFieldType.NUMBER
    default = "0"

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, value: str | float | int):
        if isinstance(value, float) or isinstance(value, int):
            value = str(value)
        if not isinstance(value, str):
            raise TypeError(f"value {type(value)} not supported")
        if re.match(r"^[0-9\.]*$", value) is None:
            raise ValueError(f"value must be a number: {value}")
        self._value = value


class LabelField(SubBaseField):
    field_type = SubFieldType.LABEL


class TextField(SubBaseField):
    field_type = SubFieldType.TEXT


class SystemDefinedField(SubBaseField, SystemDefined):
    field_type = SubFieldType.SYSTEM_DEFINED
    required_fields = ['option_layers', 'value_system']

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        SystemDefined.__init__(self, json_data)


from chemotion_api.elements.abstract_element import AbstractElement
from chemotion_api.elements.workflow_interface import WorkflowInterface
from chemotion_api.elements.workflow_manager import Workflow
from chemotion_api.utils import parse_generic_object_json, clean_generic_object_json, parse_generic_object_layer_json


class GenericElement(AbstractElement, WorkflowInterface):
    """
    By default Chemotion ELN contains five elements: samples,
    reactions, wellplates, screens, and research plans. Optionally,
    generic elements can be added by the administration.

    Usage::

    >>> from chemotion_api import Instance
    >>> from chemotion_api.collection import Collection
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the Element 'plan' with ID 1
    >>> elem = instance.get_generic_by_name('plan', 1)
    >>> elem.save()



    """

    def _set_json_data(self, json_data: dict):
        super()._set_json_data(json_data)

    def save_url(self) -> str:
        if not self.is_new:
            return "/api/v1/generic_elements/{}".format(self.id)
        return "/api/v1/generic_elements"

    def _parse_properties(self) -> dict:
        self.workflow = Workflow(self.json_data['properties_release'])
        self.workflow.init_wf(self.json_data)
        data = parse_generic_object_json(self.json_data)
        self._properties_mapping = data['obj_mapping']
        self._validator = data['validator']
        return data['values']

    def _clean_properties_data(self, serialize_data: dict | None = None) -> dict:
        clean_generic_object_json(self.json_data, self._properties, self._properties_mapping)
        return self.json_data

    def wf_tree(self) -> dict[str,list]:
        wf_tree = self.workflow.wf_tree

        def reg_wf_tree(wf_tree):
            return {wf_tree['layer']['label']: [reg_wf_tree(x) for x in wf_tree['next'].values()]}

        res = {}
        for node in wf_tree.values():
            res |= reg_wf_tree(node)
        return res

    def wf_current_workflow(self) -> list[list[str]]:
        current_wf = self.workflow.workflows
        wfs = []
        for wf in current_wf:
            wfs.append([x['label'] for x in wf['layers']])
        return wfs

    def wf_next_options(self, flat: bool = True) -> list:
        return self.workflow.next_options(flat)

    def wf_next(self, layer_id: None | str = None, wf_idx: int = 0, wf_id: str = None, **kwargs):
        key, layer = self.workflow.next(
            json_data=self.json_data,
            layer_id=layer_id,
            wf_idx=wf_idx,
            wf_id=wf_id,
            **kwargs
        )
        if layer is None:
            raise ValueError('No next step has been found. This could be that the workflow is completed or that the '
                             'arguments are not correct!')
        parse_generic_object_layer_json(key, layer, self._properties, self._properties_mapping)

    def wf_add_layer(self, layer_label: str, pos: int | None = None):
        key, layer = self.workflow.add_layer(
            json_data=self.json_data,
            layer_label=layer_label,
            pos=pos
        )

        if layer is None:
            raise ValueError(f'Layer {layer_label} has been found.')

        parse_generic_object_layer_json(key, layer, self._properties, self._properties_mapping)

    def on_properties_change(self, dict_obj, name, value):
        dict_obj[-1][name[-1]] = self._validator.validate(name[0], name[1], value[-1])
        return False



import uuid
from enum import Enum
from string import ascii_lowercase
from typing import Optional, Literal, Generator

from typing_extensions import Self

from chemotion_api.elements.reaction import Reaction
from chemotion_api.elements.sample import Sample
from chemotion_api.elements.abstract_element import AbstractElement


class BodyElements(Enum):
    richtext = 'richtext'
    ketcher = 'ketcher'
    table = 'table'
    image = 'image'
    sample = 'sample'
    reaction = 'reaction'


class Table(dict):
    """
    Table is a dict subclass. It contains additional table management methods

    Usage::

    >>> from chemotion_api import Instance
    >>> from chemotion_api.collection import Collection
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the reaction with ID 1
    >>> plan = instance.get_research_plan(1)
    >>> # Add a new Table
    >>> table = plan.add_table()
    >>> # Add a new column Name, age
    >>> name_col_id, age_col_id = table.add_columns("Name", "Age")
    >>> table.add_columns("Is Developer")
    >>> table.add_row('Mr. X', False, **{age_col_id: 40}).add_row('Mr. Y')
    >>> table.set_entry(1, age_col_id, 30)
    >>> age_average = sum(table.get_column(age_col_id)) / len(table)
    >>> print(f'Average age {age_average}')
    >>> plan.save()
    >>> plan_test = instance.get_research_plan(plan.id)
    >>> table = plan_test.body[0]['value']
    >>> age_average = sum(table.get_column(age_col_id)) / len(table)
    >>> assert age_average == 35
    >>> assert table.get_column(name_col_id) == ['Mr. X', 'Mr. Y']
    """

    def __init__(self, id: str, value: dict, start_collapsed: bool = False):
        value = {
            'columns': value.get('columns', []),
            'rows': value.get('rows', [])
        }
        super().__init__(value=value, type=BodyElements.table.value, id=id, startCollapsed=start_collapsed)

    @property
    def id(self) -> str:
        return self['id']

    def _get_col_id(self):
        for l in ascii_lowercase:
            if len([x for x in self['value']['columns'] if x['colId'] == l]) == 0:
                return l

    def column_id_by_label(self, label: str) -> str:
        """
        Finds a column ID of a column with a give label

        :param label: Label of the column

        :return: Column id
        """
        return next(self.column_ids_by_labels(label))

    def column_ids_by_labels(self, *labels: str) -> Generator[str, None, None]:
        """
        Finds all column IDs of columns with a given labels

        :param labels: Labels of the column

        :return: Column id
        """

        for l in labels:
            yield next(x['field'] for x in self['value']['columns'] if x['headerName'] == l)

    def add_columns(self, *labels: str) -> list[str]:
        """
        Adds one or more  columns to the table. It adds a column for
        each label give as an argument

        :param labels: labels of columns to be added
        :return: list of the column IDs
        """

        ids = []
        for label in labels:
            id_letter = self._get_col_id()
            self['value']['columns'].append({
                'headerName': label,
                'field': id_letter,
                'colId': id_letter
            })
            ids.append(id_letter)
            for row in self['value']['rows']:
                if id_letter not in row:
                    row[id_letter] = []

        return ids

    def add_row(self, *entries, **placed_entries) -> Self:
        """
        Adds row to the table. You can create the table empty
        or filled. To fill the fields, the values can be entered
        as named arguments or as arguments without a keyword.
        For named arguments, you must use the column id

        :param entries: non-keyworded arguments entries
        :param placed_entries:  keyworded arguments entries
        :return: Self
        """

        new_row = {}
        entries = list(entries)
        for idx, colId in enumerate(x['field'] for x in self['value']['columns']):
            if colId in placed_entries:
                new_row[colId] = placed_entries[colId]
            elif len(entries) != 0:
                new_row[colId] = entries.pop(0)
            else:
                new_row[colId] = ''

        self['value']['rows'].append(new_row)
        return self

    def set_entry(self, row: int, column_id: str, value: any) -> Self:
        """
        Sets an entry to the table

        :param row: rwo idx starts with 0
        :param column_id: the column ID
        :param value: of the Entry

        :return self
        """

        self['value']['rows'][row][column_id] = value
        return self

    def get_column(self, column_id: str) -> list:
        """
        Adds an entry to the table

        :param column_id: the column ID
        :return: a list of all values of one column
        """

        result = []
        for row in self['value']['rows']:
            result.append(row[column_id])
        return result

    def __len__(self):
        return len(self['value']['rows'])


class RichText(dict):
    class HeaderType(Enum):
        H1 = 1
        H2 = 2
        H3 = 3

    def __init__(self, id: str, value: Optional[dict] = None):
        if value is None:
            value = {"ops": []}
        value = {"ops": value.get('ops', [])}

        super().__init__(**{
            'id': id,
            'type': BodyElements.richtext.value,
            'value': value
        })

    @property
    def id(self) -> str:
        return self['id']

    @property
    def value(self) -> list[dict]:
        return self['value']['ops']

    def add_text(self, text: str,
                 bold: Optional[bool] = None,
                 underline: Optional[bool] = None,
                 italic: Optional[bool] = None,
                 list: Optional[Literal['bullet', 'ordered']] = None,
                 header: Optional[HeaderType] = None) -> Self:
        attributes = {}
        if bold is not None:
            attributes['bold'] = bold
        if underline is not None:
            attributes['underline'] = underline
        if italic is not None:
            attributes['italic'] = italic
        if list is not None and list in ['bullet', 'ordered']:
            attributes['list'] = list
        if header is not None:
            attributes['header'] = header.value

        if len(attributes) == 0:
            self.value.append({"insert": text})
        else:
            self.value.append({"insert": text, 'attributes': attributes})

        return self


class ResearchPlan(AbstractElement):
    """
    A chemotion Research Plan object.
    It extends the :class:`chemotion_api.elements.abstract_element.AbstractElement`

    Usage::

    >>> from chemotion_api import Instance
    >>> from chemotion_api.collection import Collection
    >>> from chemotion_api.elements.research_plan import Table, RichText
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> # Get the reaction with ID 1
    >>> plan = instance.get_research_plan(1)
    >>> # Add a table
    >>> table: Table = plan.add_table()
    >>> table.add_columns('Name')
    >>> table.add_row('Martin')
    >>> # Add a simple Text
    >>> plan.add_richtext("Hallo")
    >>> # Add more advanced text
    >>> rt = plan.add_richtext()
    >>> # Add header to the text
    >>> rt.add_text('Header 1\\n', header=RichText.HeaderType.H1).add_text('Header 2\\n', header=RichText.HeaderType.H2)
    >>> # Add Bold,  Italic & underline text
    >>> rt.add_text('Bold,  Italic & underline\\n', bold=True, italic=True, underline=True)
    >>> # Add ordered and  list
    >>> rt.add_text('a\\nb\\nc\\n', list='ordered').add_text('a\\nb\\nc\\n', list='bullet')
    >>> plan.save()
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def body(self) -> list[dict]:
        """
        Body is list which contains the same values as the Properties.
        Each entry has a type and a value filed.
        The type can be one of:  richtext, ketcher, table, image, sample or reaction

        :return: List of body entries
        """
        return self.properties['body']

    def _set_json_data(self, json_data):
        super()._set_json_data(json_data)

    def _parse_body_element(self, elem: dict):
        value = ''
        if elem.get('type') == BodyElements.richtext.value:
            return self._parse_text(elem)
        elif elem.get('type') == BodyElements.ketcher.value:
            value = self._parse_ketcher(elem.get('value'))
        elif elem.get('type') == BodyElements.table.value:
            return self._parse_table(elem)
        elif elem.get('type') == BodyElements.image.value:
            value = self._parse_image(elem.get('value'))
        elif elem.get('type') == BodyElements.sample.value:
            value = self._parse_sample(elem.get('value'))
        elif elem.get('type') == BodyElements.reaction.value:
            value = self._parse_reaction(elem.get('value'))

        return {
            'id': elem.get('id'),
            'type': elem.get('type'),
            'value': value
        }

    def _parse_sample(self, value):
        if value is None or value.get('sample_id') is None:
            return None
        return Sample(self._generic_segments,
                      self._session,
                      id=value.get('sample_id'), element_type='sample')

    def _parse_reaction(self, value):
        if value is None or value.get('reaction_id') is None:
            return None
        return Reaction(self._generic_segments,
                        self._session,
                        id=value.get('reaction_id'), element_type='reaction')

    def _parse_text(self, value):
        return RichText(value['id'], value['value'])

    def _parse_image(self, value):
        if value is None: return None
        try:
            res = self.attachments.load_attachment(identifier=value.get('public_name'))
        except ValueError:
            return None
        return res

    def _parse_table(self, value):
        return Table(value['id'], value['value'], value.get('startCollapsed', False))

    def _parse_ketcher(self, value):
        return value

    def _parse_properties(self) -> dict:
        body = self.json_data.get('body')
        return {
            'body': [self._parse_body_element(x) for x in body]
        }

    def _clean_properties_data(self, serialize_data: dict | None = None) -> dict:
        self.json_data['body'] = []
        for elem in self.body:
            self.json_data['body'].append(self._reparse_body_element(elem))

        return self.json_data

    def add_richtext(self, text: Optional[str] = None, at_idx: Optional[int] = None) -> RichText:
        """
        Adds a text block to the Research plan

        :param text: to be added

        :return: body container
        """
        body_obj: RichText = self._add_new_element(BodyElements.richtext, at_idx)
        if text is not None:
            body_obj.add_text(text)
        return body_obj

    def add_image(self, image_path: str, at_idx: Optional[int] = None) -> dict:
        """
        Adds an image to the Research plan

        :param image_path: file path to image

        :return: body container
        """

        if not image_path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif')):
            raise ValueError('File is not a image!')
        file_obj = self.attachments.add_file(image_path)
        file_obj['is_image_field'] = True
        file_obj['ancestor'] = None
        body_obj = self._add_new_element(BodyElements.image, at_idx)

        body_obj['value'] = file_obj
        return body_obj

    def add_table(self, at_idx: Optional[int] = None) -> Table:
        """
        Adds a new table to the Research plan

        :return: Table object
        """

        body_obj = self._add_new_element(BodyElements.table, at_idx)
        return body_obj

    def _add_new_element(self, element_type: BodyElements, at_idx: Optional[int]):
        body_elem = {
            'id': uuid.uuid4().__str__(),
            'type': element_type.name,
            'value': self._default_body_element(element_type)
        }

        new_element = self._parse_body_element(body_elem)
        if at_idx is None:
            self.body.append(new_element)
        else:
            self.body.insert(at_idx, new_element)
        return new_element

    def _reparse_body_element(self, elem: dict):
        if elem.get('type') == BodyElements.richtext.value:
            value = {'value': self._reparse_text(elem.get('value'))}
        elif elem.get('type') == BodyElements.ketcher.value:
            value = {'value': self._reparse_ketcher(elem.get('value'))}
        elif elem.get('type') == BodyElements.table.value:
            value = self._reparse_table(elem)
        elif elem.get('type') == BodyElements.image.value:
            value = {'value': self._reparse_image(elem.get('value'))}
        elif elem.get('type') == BodyElements.sample.value:
            value = {'value': self._reparse_sample(elem.get('value'))}
        elif elem.get('type') == BodyElements.reaction.value:
            value = {'value': self._reparse_reaction(elem.get('value'))}
        else:
            value = {'value': ''}

        elem_data = {
            'id': elem.get('id'),
            'type': elem.get('type'),
        } | value

        if elem.get('type') == 'richtext':
            elem_data['title'] = 'Text'
        return elem_data

    def _reparse_sample(self, value: Sample | None):
        if value is None:
            return {'sample_id': None}
        return {'sample_id': value.id}

    def _reparse_reaction(self, value: Reaction | None):
        if value is None:
            return {'reaction_id': None}
        return {'reaction_id': value.id}

    def _reparse_text(self, value):
        return value

    def _reparse_image(self, value):
        return {
            'public_name': value['identifier'],
            'file_name': value['filename']
        }

    def _reparse_table(self, value):
        return dict(value)

    def _reparse_ketcher(self, value):
        return value

    @staticmethod
    def _default_body_element(element_type: BodyElements) -> dict:
        if element_type == BodyElements.richtext:
            return {'ops': []}
        if element_type == BodyElements.ketcher:
            return {
                'svg_file': None,
                'thumb_svg': None
            }
        if element_type == BodyElements.table:
            return {
                'columns': [],
                'rows': []

            }
        if element_type == BodyElements.image:
            return {
                'file_name': None,
                'public_name': None,
                'zoom': None
            }
        if element_type == BodyElements.sample:
            return {
                'sample_id': None
            }
        if element_type == BodyElements.reaction:
            return {
                'reaction_id': None
            }
        raise ValueError(f"{element_type} not exists!")



import json
import os
import typing
from datetime import datetime
from urllib.parse import urlparse
from enum import Enum
from random import random
from typing import TypeVar, Optional, Self, Generator

from chemotion_api.elements import Wellplate, Sample, Reaction, GenericElement, ResearchPlan
from chemotion_api.user import Person, Group

from chemotion_api.elements.sample import MoleculeManager
from chemotion_api.connection import Connection

from chemotion_api.element_manager import ElementManager
from chemotion_api.elements import ElementSet, AbstractElement

TAbstractCollection = TypeVar("TAbstractCollection", bound="AbstractCollection")
TRootCollection = TypeVar("TRootCollection", bound="RootCollection")


def _col_path_join(path_part: str, *path_parts: str) -> str:
    path_parts_list = list(path_parts)
    path_parts_list.reverse()
    for i, x in enumerate(path_parts):
        if x.startswith('/'):
            if i == 0:
                return _col_path_join(x)
            return _col_path_join(x, *path_parts[-i:])
    path_parts_str = '/'.join(path_parts)
    return f'{path_part}/{path_parts_str}'.replace('//', '/').replace('//', '/')


class SyncPermission(Enum):
    Read = 0
    Write = 1
    Share = 2
    Delete = 3
    ImportElements = 4
    PassOwnership = 5


class AbstractCollection:

    def __init__(self):
        self.is_sync = False
        self._children: list[Self] = []
        self._label: str | None = None
        self._parent: Self | None = None
        self._id: int | None = None

    @property
    def children(self) -> list[Self]:
        """
        A list of all children collection

        :return: A list of all children collection
        """
        return self._children

    @property
    def id(self) -> int | None:
        """
        The chemotion DB ID

        :return: The chemotion DB ID
        """
        return self._id

    @id.setter
    def id(self, id: int):
        """
        The chemotion DB ID
        """
        self._id = id

    @property
    def label(self) -> str | None:
        """
        The collection label

        :return: The collection label
        """
        return self._label

    @label.setter
    def label(self, label: str):
        """
        The collection label
        """
        self._label = label

    def prepare_label(self, case_insensitive: bool) -> str:
        """
        Prepares the label of a collection

        :param case_insensitive: If true it returns teh lower case label

        :return: The collection's label
        """
        if case_insensitive:
            return self.label.lower()
        return self.label

    def __str__(self) -> str:
        return self.get_path()

    def __iter__(self) -> Generator[tuple[str, list[Self]], None, None]:
        """
        Iterate over all children

        :return: Iterator
        """
        yield 'collections', self._children

    def move(self, dest: str | list[str], case_insensitive: bool = False):
        raise NotImplementedError('Cannot not move this Collection')

    def _set_children(self, new_children: list[dict]):
        if new_children is None:
            self._children = []
            return
        ids = []
        for child in new_children:
            ids.append(child['id'])
            child_obj: Self | None = next((x for x in self._children if x.id == child['id']), None)
            if child_obj is None:
                self._children.append(Collection(child))
            else:
                child_obj.set_json(child)
        self._children = [child_obj for child_obj in self._children if child_obj.id in ids]
        self._update_relations()

    def _update_relations(self):
        for child in self._children:
            child._parent = self
            child.is_sync = self.is_sync
            child._update_relations()

    def to_json(self) -> dict:
        """
        Generates a dict containing all list of all children collection

        :return: list of children as savable JSON
        """
        return {'children': [x.to_json() for x in self._children]}

    def find(self, label: str | None = None, **kwargs) -> list[typing.Union['AbstractCollection', 'Collection']]:
        """
        Finds a collection by its label or other properties. Path is not needed.

        :param label: String or None.
        :param kwargs: Other properties used to search Collections

        :return: All found Collections
        """

        results: list[Collection | AbstractCollection] = []
        if label is not None: kwargs['label'] = label
        hit = True
        for (key, val) in kwargs.items():
            if getattr(self, key) != val:
                hit = False
                break
        if hit:
            results.append(self)
        for x in self._children:
            results += x.find(**kwargs)
        return results

    def get_path(self) -> str:
        """
        Generates the absolut path of collection

        :return: Absolut path of a collection
        """

        abs_path = []
        col = self
        while col._parent is not None:
            abs_path.append(col.label)
            col = col._parent
        abs_path.append('')
        abs_path.reverse()
        return '/'.join(abs_path)

    def get_root(self) -> TRootCollection:
        """
        Get root Collection

        :return: Instance of the RootCollection
        """

        col = self
        while col._parent is not None:
            col = col._parent
        return col

    def get_collection(self, col_path: str | list[str], case_insensitive: bool = False) -> 'AbstractCollection':
        """
        Gets a collection a path points to.

        :param col_path: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account

        :return: Collection the path points to
        """

        abs_path = self.get_path()
        if col_path.__class__ is not str:
            col_path = '/'.join(col_path)
        return self.get_root().get_collection(_col_path_join(abs_path, col_path), case_insensitive)

    def get_or_create_collection(self, col_path: str | list[str],
                                 case_insensitive: bool = False) -> TAbstractCollection:
        """
        Gets or creates a collection the col_path path points to. If the collection needs to be created
        the collection is saved right away.

        :param col_path: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account

        :return: Collection the path points to
        """

        try:
            return self.get_collection(col_path, case_insensitive)
        except:
            new_col = self.add_collection(col_path)
            root = self.get_root()
            new_path = new_col.get_path()
            root.save()
            return root.get_collection(new_path)

    def add_collection(self, col_path: str | list[str]):
        """
        Creates a new collection at the path_to_new path. The collection is not saved.
        Please reload the collection after saving it.

        :param col_path: is either a string based on a unix path or a list

        :return: new Collection (not saved)

        :raises: Exception if collection is synced. This is not yet implemented
        """

        if self.is_sync:
            raise Exception("You cannot add a collection to a synced collection!")
        raise NotImplementedError('This collection cannot add a collection')

    def get_elements_of_iri(self, element_type: str, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        Gets all elements of given IRI type.

        :param element_type: IRI string chemotion:xxx/xxx/xxx
        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet based on the IRI
        """

        root = self.get_root()
        if element_type in root._element_manager.all_types:
            o = urlparse(element_type)
            e = ElementSet(root._session, root._element_manager.all_classes.get(o.path.split('/')[1]), self.id,
                           self.is_sync)
            e.load_elements(per_page, updated_to_date, updated_from_date)
            e.set_iri_filter(element_type)
            return e
        raise ValueError(f'Could not find a element with the IRI: "{element_type}"')

    def get_samples(self, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None, product_only: bool = False) -> ElementSet:
        """
        List of all samples.

        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'
        :param product_only: If True only samples which are products in a reaction are returned

        :return: ElementSet of samples
        """

        root = self.get_root()
        e = ElementSet(root._session, root._element_manager.all_classes.get('sample'), self.id,
                       self.is_sync)
        e.load_elements(per_page, updated_to_date, updated_from_date, product_only)
        return e

    def get_reactions(self, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        List of all reactions.

        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet of reactions
        """

        root = self.get_root()
        e = ElementSet(root._session, root._element_manager.all_classes.get('reaction'), self.id,
                       self.is_sync)
        e.load_elements(per_page, updated_to_date, updated_from_date)
        return e

    def get_research_plans(self, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        List of all research plans.

        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet of research plans
        """

        root = self.get_root()
        e = ElementSet(root._session, root._element_manager.all_classes.get('research_plan'), self.id,
                       self.is_sync)
        e.load_elements(per_page, updated_to_date, updated_from_date)
        return e

    def get_wellplates(self, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        List of all well plates.

        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet of well plates
        """

        root = self.get_root()
        e = ElementSet(root._session, root._element_manager.all_classes.get('wellplate'), self.id,
                       self.is_sync)
        e.load_elements(per_page, updated_to_date, updated_from_date)
        return e

    def get_generics_by_name(self, name: str, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        List of all generic elements of a given name.

        :param name: generic element's name
        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet of generic elements of a given name
        """

        root = self.get_root()
        elem = root._element_manager.all_classes.get(name)
        if elem is None:
            raise ValueError(f'Could not find a generic element under the name: "{name}"')

        e = ElementSet(root._session, elem, self.id, self.is_sync)
        e.load_elements(per_page, updated_to_date, updated_from_date)
        return e

    def get_generics_by_label(self, label: str, per_page: Optional[int] = 10, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None) -> ElementSet:
        """
        List of all generic elements of a given label.

        :param label: generic element's label
        :param per_page: How many Elements per page
        :param updated_to_date: Set date time to fetch only elements which are updated before 'updated_to_date'
        :param updated_from_date: Set date time to fetch only elements which are updated after 'updated_from_date'

        :return: ElementSet of generic elements of a given name
        """

        root = self.get_root()
        for (elem_name, elem) in root._element_manager.all_classes.items():
            if elem['label'] == label:
                return self.get_generics_by_name(elem_name, per_page, updated_to_date, updated_from_date)
        raise ValueError(f'Could not find a generic element with the label: "{label}"')


class AbstractEditableCollection(AbstractCollection):

    def new_sample(self) -> Sample:
        """
        Creates a new sample in this collection.

        :return: New (unsaved) Sample
        """
        return typing.cast(Sample, self._create_new_element('sample'))

    def new_sample_smiles(self, smiles_code: str) -> Sample:
        """
        Creates a new sample with a molecule generated with a smiles code in this collection.

        :return: New (unsaved) Sample
        """
        sample = self._create_new_element('sample')
        mol = MoleculeManager(self.get_root()._session).create_molecule_by_smiles(smiles_code)
        sample.molecule = mol
        return typing.cast(Sample, sample)

    def new_solvent(self, name) -> Sample:
        """
        Creates a new solvent in this collection.

        :return: New (unsaved) solvent as Sample
        """
        root = self.get_root()
        new_json = root._element_manager.build_solvent_sample(name, self.id)
        e = ElementSet(root._session, root._element_manager.all_classes.get('sample'), self.id,
                       self.is_sync)
        return typing.cast(Sample, e.new_element(new_json, is_solvent=True))

    def new_reaction(self) -> Reaction:
        """
        Creates a new reaction in this collection.

        :return: New (unsaved) Reaction
        """
        return typing.cast(Reaction, self._create_new_element('reaction'))

    def new_research_plan(self) -> ResearchPlan:
        """
        Creates a new research plan in this collection.

        :return: New (unsaved) ResearchPlan
        """
        return typing.cast(ResearchPlan, self._create_new_element('research_plan'))

    def new_wellplate(self) -> Wellplate:
        """
        Creates a new well plate in this collection.

        :return: New (unsaved) Wellplate
        """

        return typing.cast(Wellplate, self._create_new_element('wellplate'))

    def new_generic(self, type_name: str) -> GenericElement:
        """
        Creates a new element of given type name in this collection.

        :param type_name: Generic element's type name

        :return: New (unsaved) GenericElement
        """

        return typing.cast(GenericElement, self._create_new_element(type_name))

    def new_generic_by_label(self, label: str) -> GenericElement:
        """
        Creates a new element of given label in this collection.

        :param label: Generic element's label

        :return: GenericElement
        """

        for (elem_name, elem) in self.get_root()._element_manager.all_classes.items():
            if elem['label'] == label:
                return typing.cast(GenericElement, self._create_new_element(elem_name))
        raise ValueError(f'Could not find a generic element with the label: "{label}"')

    def new_element_by_iri(self, element_type: str) -> AbstractElement:
        """
        Creates a new element of given IRI type. However, if the type is generic the generated element is of the newest type.

        :param element_type: IRI string chemotion:xxx/xxx/xxx

        :return: AbstractElement
        """

        if element_type in self.get_root()._element_manager.all_types:
            o = urlparse(element_type)
            return self._create_new_element(o.path.split('/')[1])
        raise ValueError(f'Could not find a generic element with the label: "{element_type}"')

    def _create_new_element(self, type_name) -> AbstractElement:
        root = self.get_root()
        new_json = root._element_manager.build_new(type_name, self.id)

        e = ElementSet(root._session, root._element_manager.all_classes.get(type_name), self.id,
                       self.is_sync)
        return e.new_element(new_json)


class Collection(AbstractEditableCollection):
    """
    The collection object represents the standard collection owned by the current user
    """

    def __init__(self, collection_json: dict = None):
        super().__init__()
        self.permission_level: int | None = None
        self.reaction_detail_level: int | None = None
        self.sample_detail_level: int | None = None
        self.screen_detail_level: int | None = None
        self.wellplate_detail_level: int | None = None
        self.element_detail_level: int | None = None
        self.sync_collections_users: dict | None = None
        self.is_locked: bool | None = None
        self.is_shared: bool | None = None
        self.is_synchronized: bool | None = None
        self.set_json(collection_json)

    collection_json: dict = None

    def set_json(self, collection_json):
        """
        Set the json data received from the server

        :param collection_json: JSON object representing the collection
        """
        if collection_json is None:
            collection_json = self._get_new_json()

        self.collection_json = collection_json
        self._set_children(collection_json.get('children', []))
        if 'children' in collection_json: del collection_json['children']

        for (key, val) in collection_json.items():
            if hasattr(self, key):
                setattr(self, key, val)

    def _get_new_json(self):
        return {
            "id": random(),
            "label": 'New Collection',
            "isNew": True
        }

    def __iter__(self):
        for (key, val) in self.collection_json.items():
            if hasattr(self, key):
                val = getattr(self, key)
            yield (key, val)

    def to_json(self):
        """
        Generates a dict containing all list of all children collection

        :return: list of children as savable JSON
        """
        as_dict = dict(self)
        return super().to_json() | as_dict

    def move(self, dest: str | list[str], case_insensitive: bool = False):
        """
        Moves this collection to a dest path.

        :param dest: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account

        :return: Collection the path points to
        """

        abs_path = self.get_path()
        dest = _col_path_join(os.path.dirname(abs_path), dest)
        self.get_root().move(abs_path, dest, case_insensitive)

    def delete(self):
        """
        Deletes this collection.
        """
        abs_path = self.get_path()
        self.get_root().delete(abs_path)

    def add_collection(self, name: str):
        """
        Creates a new child collection. The collection is not saved.
        Please reload the collection after saving it.

        :param name: new collection`s name

        :return: new Collection (not saved)
        """
        abs_path = _col_path_join(self.get_path(), name)
        return self.get_root().add_collection(abs_path)

    def share(self, permission_level: SyncPermission, *users: Person | Group):
        """
        To synchronize a collection with a set of users or groups

        :param permission_level: Element from SyncPermission
        :param users: a list of users or groups
        """
        data = {
            "collection_attributes": {
                "permission_level": permission_level.value,
                "sample_detail_level": 10,
                "reaction_detail_level": 10,
                "wellplate_detail_level": 10,
                "screen_detail_level": 10,
                "element_detail_level": 10
            },
            "user_ids": [
                {'label': f'{user.name} ({user.initials} - {user.user_type})',
                 'name': user.name,
                 'value': user.id} for user in users if user.is_group() or user.is_person()
            ],
            "id": self.id
        }

        res = self.get_root()._session.post('/api/v1/syncCollections/', data=data)
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")


class RootSyncCollection(AbstractCollection):
    """
    This can be used amost like the root collection object.
    However, you cannot move, delete or create its child collections
    """

    def __init__(self, session: Connection, element_manager: ElementManager):
        super().__init__()
        self.is_sync = True
        self._session = session
        self._element_manager = element_manager
        self.label = 'sync_root'

    def to_json(self):
        as_dict = dict(self)
        return super().to_json() | as_dict

    def move(self, *args, **kwargs):
        raise Exception("You cannot move a synced collection collection!")

    def delete(self, *args, **kwargs):
        raise Exception("You cannot delete a synced collection collection!")


class RootCollection(AbstractCollection):
    """
    The root collection is the foundation of working with collections. It serves with methods to navigate through the
    collections and creating new collections



    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> rc = instance.get_root_collection(True)
    >>> # Navigate to collection 'Team One' in 'Projects'
    >>> col = rc.get_collection('/Projects/Team One')
    >>> # Create collection 'P 123' in '/Project/Team One'
    >>> col.get_or_create_collection('./P 123')
    >>> # Move collection P 122 into collection '/Done Projects'
    >>> col.get_collection('./P 122').move('/Done Projects')
    >>> for sample in col.get_samples():
    >>>     pass # Do something
    """

    def __init__(self, session: Connection):
        super().__init__()
        self._sync_root: RootSyncCollection | None = None
        self._all: dict | None = None
        self._element_manager: ElementManager
        self._session = session
        self.label = 'root'
        self._deleted_ids = []
        self._id = None

    @property
    def id(self) -> int:
        """
        The 'All' collection ID

        :return: Ihe ID of the 'All' collection
        """

        if self._id is None:
            self.load_collection()
        return self._id

    @property
    def sync_root(self) -> RootSyncCollection:
        """
        The RootSyncCollection collection can be used to access all collection synced with you.

        :return: instance of RootSyncCollection
        """

        if self._sync_root is None:
            self.load_sync_collection()
        return self._sync_root

    @property
    def all(self) -> dict[str,dict]:
        """
        Returns all information of the 'All' collections

        :return: information of the 'All' collection
        """

        if self._all is None:
            self.load_collection()
        return self._all

    def set_element_manager(self, element_manager: ElementManager):
        """
        Set the ElementManager class

        :param element_manager: instance of ElementManager
        """

        self._element_manager = element_manager

    def load_collection(self) -> dict[str,dict]:
        """
        Load collection information

        :return: A collection off all collection information
        """

        collection_url = '/api/v1/collections/roots.json'

        res = self._session.get(collection_url)

        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        collections = json.loads(res.content)
        self._all = self._load_all_collection()['collection']
        self._id = self._all['id']
        self._set_children(collections['collections'])
        return self._all

    def load_sync_collection(self) -> RootSyncCollection:
        """
        Loads the RootSyncCollection. The RootSyncCollection collection can be used to access all collection synced with you.

        :return: instance of RootSyncCollection
        """

        collection_url = '/api/v1/syncCollections/sync_remote_roots'

        res = self._session.get(collection_url)

        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        collections = json.loads(res.content)
        self._sync_root = RootSyncCollection(self._session, self._element_manager)
        self._sync_root._set_children(collections['syncCollections'])
        return self._sync_root

    def save(self):
        """
        Saves ALL updates of done to owned (not synced) collection.
        """

        collection_url = '/api/v1/collections'
        payload = self.to_json()
        payload['deleted_ids'] = self._deleted_ids
        res = self._session.patch(collection_url,
                                  data=payload)
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        self.load_collection()

    def get_collection(self, col_path: str | list[str], case_insensitive: bool = False) -> 'AbstractCollection':
        """
        Gets a collection a path points to.

        :param col_path: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account

        :return: Collection the path points to
        """

        col_path = self._prepare_path(col_path)
        current_pos = self
        for col_label in self._prepare_path(col_path):
            current_pos = next((x for x in current_pos.children if
                                x.prepare_label(case_insensitive) == col_label), None)
            if current_pos is None:
                raise ModuleNotFoundError("'{}' Collection Not Found".format('/'.join(col_path)))
        return current_pos

    def move(self, src: str | list[str], dest: str | list[str], case_insensitive: bool = False):
        """
        Moves a collection from a src path to a dest path.

        :param src: is either a string based on a unix path or a list
        :param dest: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account

        :return: Collection the path points to
        """

        prepared_src = self._prepare_path(src, case_insensitive)
        src_col = self.get_collection(prepared_src, case_insensitive)

        idx = next((i for (i, x) in enumerate(src_col._parent.children) if
                    x.prepare_label(case_insensitive) == prepared_src[-1]), None)
        dest_col = self.get_collection(dest, case_insensitive)

        src_col._parent.children.pop(idx)
        dest_col.children.append(src_col)
        self._update_relations()

    def delete(self, src: str | list[str], case_insensitive: bool = False):
        """
        Deletes a collection at the src path.

        :param src: is either a string based on a unix path or a list
        :param case_insensitive: If true the cases of path are taken into account
        """

        prepared_src = self._prepare_path(src, case_insensitive)
        src_col = self.get_collection(prepared_src)
        idx = next((i for (i, x) in enumerate(src_col._parent.children) if x.label == prepared_src[-1]), None)

        src_col._parent.children.pop(idx)
        src_col._parent = None
        self._deleted_ids.append(src_col.id)

    def add_collection(self, path_to_new: str | list[str]):
        """
        Creates a new collection at the path_to_new path. The collection is not saved.
        Please reload the collection after saving it.

        :param path_to_new: is either a string based on a unix path or a list

        :return: new Collection (not saved)
        """

        prepared_src = self._prepare_path(path_to_new)
        src_col = self.get_collection(prepared_src[:-1])
        c = Collection()
        c.label = prepared_src[-1]
        src_col.children.append(c)
        self._update_relations()
        return c

    def to_json(self) -> dict[str, list]:
        """
        Generates a dict containing all list of all children collection

        :return: list of children as savable JSON
        """

        return {'collections': super().to_json()['children']}

    def _prepare_path(self, col_path: str | list[str], case_insensitive: bool = False) -> list[str]:
        if type(col_path) == str:
            if not col_path.startswith('/'): col_path = '/' + col_path
            col_path = [x for x in col_path.strip('/').split('/') if x != '']
            if case_insensitive:
                col_path = [x.lower() for x in col_path]

        return col_path

    def _load_all_collection(self):
        collection_url = '/api/v1/collections/all'

        res = self._session.get(collection_url)

        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        return json.loads(res.content)



import uuid
from chemotion_api.connection import Connection

class GenericSegments:
    def __init__(self, session: Connection):
        self._session = session
        self._segments = None

    @property
    def all_classes(self):
        if self._segments is None:
            get_url = "/api/v1/segments/klasses.json"
            res = self._session.get(get_url)
            if res.status_code != 200:
                raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
            self._segments = res.json().get('klass', [])

        return self._segments

    @classmethod
    def new_session(cls, base_segment: dict) -> dict:
        a = {
            'segment_klass_id': base_segment['id'],
            'klassType': 'Segment',
            'is_new': True,
            'id': uuid.uuid4().__str__()
        }
        a['properties'] = base_segment.get('properties_release', {}).copy()
        a['properties_release'] = base_segment.get('properties_release', {}).copy()

        return a


from __future__ import annotations

from chemotion_api.elements.utils import QuantityUnit, GasType, GasCalculations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from chemotion_api.elements import Sample


def calculate_equivalent_for_gas_product(sample: Sample, ppm: float, temperature_in_k: float) -> None | float:
    reaction = sample.reaction

    ref_material = reaction.get_feedstock()
    if ref_material is None or not sample.vessel_size:
        return None
    f_purity = ref_material.properties['purity']
    try:
        return ppm * GasCalculations.DEFAULT_TEMPERATURE_IN_KELVIN / (
                f_purity * temperature_in_k * GasCalculations.PARTS_PER_MILLION_FACTOR)
    except TypeError:
        return None


def get_equivalent(sample: Sample, mol: float, ref_mol: float, ppm: float, temperature_in_k: float):
    role = sample.role_in_reaction
    if sample.properties['reference']:
        return 1
    if role == 'products':
        if sample.gas_type == GasType.GAS:
            return calculate_equivalent_for_gas_product(sample, ppm, temperature_in_k)

        try:
            stoichiometry_coefficient = (sample.properties['coefficient'] or 1.0) / (
                        sample.reaction.reference_sample.properties['coefficient'] or 1.0)
            return mol / ref_mol / stoichiometry_coefficient
        except ZeroDivisionError or AttributeError:
            return None
    elif role in ['starting_materials', 'reactants']:
        try:
            return mol / ref_mol
        except ZeroDivisionError or AttributeError:
            return None
    return None



import json
import os
import uuid


def init_container():
    json_path = os.path.join(os.path.dirname(__file__), 'container.json')
    f = open(json_path, 'r')
    data = json.loads(f.read())
    root = data.copy()
    root['container_type'] = 'root'
    root['id'] = uuid.uuid4().__str__()
    analyses = data.copy()
    analyses['id'] = uuid.uuid4().__str__()
    analyses['container_type'] = 'analyses'
    root['children'] = [analyses]
    f.close()
    return root





import os
import subprocess
from datetime import datetime
from .creator import walk_productive_template

import click
import requests

from .updater import update_images, update_eln_port, update_db_port, get_eln_port, get_db_port


class InstanceManager:

    def __init__(self, df=None):
        if df is None:
            df = os.getcwd()
        if os.path.isdir(df):
            self._df_name = 'docker-compose.yml'
            self._df_dir = df
        else:
            self._df_dir = os.path.dirname(df)
            self._df_name = os.path.basename(df)

    @property
    def eln_port(self):
        df = self._check_file()
        return get_eln_port(df)

    @property
    def db_port(self):
        df = self._check_file()
        return get_db_port(df)

    def update(self, version=None):
        df = self._check_file()
        images = self.get_image_of_versions(version)
        update_images(df, **images)

    def update_port(self, eln_port):
        df = self._check_file()
        update_eln_port(df, eln_port)

    def update_db_port(self, postgresql_port):
        df = self._check_file()
        update_db_port(df, postgresql_port)

    def check(self):
        self._check_file()

        command = ['docker', 'compose', '-f', self._df_name, 'ps']
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, cwd=self._df_dir)
        stdout, stderr = proc.communicate()
        a = stdout.decode()
        click.echo(a)
        return [[y.strip() for y in x.split('   ') if y.strip() != ''] for x in a.split('\n')[2:]]

    def new_instance(self, version: str | None = None, url: None | str = None, postgres_port: int = 5432,
                     eln_port: int = 4000, adminer_port: int = 8080, cwd=None):
        images = self.get_image_of_versions(version)
        if cwd is None:
            cwd = self._df_dir
        cwd = os.path.abspath(cwd)
        if not os.path.isdir(cwd):
            if os.path.exists(os.path.isdir(cwd)):
                raise NotADirectoryError(f"{cwd} is not a directory")
            else:
                os.makedirs(cwd, exist_ok=True)
        if len(os.listdir(cwd)) != 0:
            raise ValueError(f'{cwd} is not empty!')
        if url is None:
            url = f'http://localhost:{eln_port}'
        for (is_file, name, content) in walk_productive_template(url, eln_port, postgres_port, adminer_port, **images):
            if is_file:
                with open(os.path.join(cwd, name), 'w+') as f:
                    f.write(content)
            else:
                os.makedirs(os.path.join(cwd, name))

        os.rename(os.path.join(cwd, 'docker-compose.yml'), os.path.join(cwd, self._df_name))

    def create_images(self):
        self._check_file()

        command = ['docker', 'compose', '-f', self._df_name, 'up', '--no-start']
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, cwd=self._df_dir)
        return proc.communicate()

    def up(self, eln: bool = False, db: bool = False, worker: bool = False, ketchersvc: bool = False,
           msconvert: bool = False, spectra: bool = False, converter: bool = False):
        return self._up(eln=eln, db=db, worker=worker, ketchersvc=ketchersvc, msconvert=msconvert, spectra=spectra,
                        converter=converter)

    def stop(self, eln: bool = False, db: bool = False, worker: bool = False, ketchersvc: bool = False,
             msconvert: bool = False, spectra: bool = False, converter: bool = False):
        return self._stop(eln=eln, db=db, worker=worker, ketchersvc=ketchersvc, msconvert=msconvert, spectra=spectra,
                          converter=converter)

    def down(self, eln: bool = False, db: bool = False, worker: bool = False, ketchersvc: bool = False,
             msconvert: bool = False, spectra: bool = False, converter: bool = False):
        return self._down(eln=eln, db=db, worker=worker, ketchersvc=ketchersvc, msconvert=msconvert, spectra=spectra,
                          converter=converter)

    def _up(self, **kwargs):
        self._check_file()

        self.create_images()

        command_suffix = []
        if not all(kwargs.values()) and any(kwargs.values()):
            command_suffix = [k for (k, v) in kwargs.items() if v]

        command = ['docker', 'compose', '-f', self._df_name, 'start'] + command_suffix
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, cwd=self._df_dir)
        stdout, stderr = proc.communicate()
        return stdout

    def _stop(self, **kwargs):
        self._stop_or_down(stop_or_down='stop', **kwargs)

    def _down(self, **kwargs):
        self._stop_or_down(stop_or_down='down', **kwargs)

    def _stop_or_down(self, stop_or_down, **kwargs):
        if stop_or_down not in ['stop', 'down']:
            raise ValueError('stop_or_down must be in stop, down')
        self._check_file()
        command_suffix = []
        if not all(kwargs.values()) and any(kwargs.values()):
            command_suffix = [k for (k, v) in kwargs.items() if v]

        command = ['docker', 'compose', '-f', self._df_name, stop_or_down] + command_suffix
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, cwd=self._df_dir)
        stdout, stderr = proc.communicate()
        return stdout

    def check_file(self) -> bool:
        try:
            self._check_file()
            return True
        except FileNotFoundError:
            return False

    def _check_file(self) -> str:
        df = str(os.path.join(self._df_dir, self._df_name))
        if not os.path.exists(df):
            raise FileNotFoundError(f"Docker-compose file dose not exist: {df}")
        return df

    @staticmethod
    def backup(pguser='postgres', port=5432, host='localhost', pgpassword='postgres', database='chemotion',
               file_path=None):
        my_env = os.environ.copy()
        my_env["PGPASSWORD"] = pgpassword
        if file_path is None:
            file_path = os.path.abspath(
                os.path.join(os.getcwd(), f'./{database}_{datetime.now().strftime("%Y_%m_%d_%H_%M_%S")}.tar'))

        command = ['pg_dump', '-p', f'{port}', '-U', pguser, '-h', host, '-F', 't', '-f', file_path, database]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            raise ChildProcessError(stderr)
        click.secho(f"Dumped {database} -> {file_path} !", fg='green', bg='white')
        return file_path

    @staticmethod
    def restore(file_path, pguser='postgres', port=5432, host='localhost', pgpassword='postgres', database='chemotion'):
        file_path = os.path.abspath(file_path)

        my_env = os.environ.copy()
        my_env["PGPASSWORD"] = pgpassword
        command = ['dropdb', '-U', pguser, '-h', host, '-p', f'{port}', '-f', '--if-exists', database]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)
        proc.communicate()
        if proc.returncode == 0:
            click.secho(f"Dropped {database}!", fg='red', bg='white')
        command = ['createdb', '-U', pguser, '-h', host, '-p', f'{port}', database]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)
        proc.communicate()
        if proc.returncode != 0:
            raise ChildProcessError(f'createdb error: {proc.returncode}')
        click.secho(f"Recreated {database}!", fg='green', bg='white')
        command = ['pg_restore', '-U', pguser, '-h', host, '-p', f'{port}', '-d', database, file_path]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)
        proc.communicate()
        if proc.returncode != 0:
            raise ChildProcessError(f'pg_restore error: {proc.returncode}')
        click.secho(f"Restored {file_path} -> {database}!", fg='green', bg='white')

    @staticmethod
    def get_versions():
        res = requests.get('https://registry.hub.docker.com/v2/repositories/ptrxyz/chemotion/tags?page_size=1024')
        return [x['name'][4:] for x in res.json()['results'] if x['name'].startswith('eln')]

    @staticmethod
    def get_image_of_versions(version: None | str = None):
        v_list = InstanceManager.get_versions()
        if version is None:
            version = v_list[0]
        res = requests.get('https://registry.hub.docker.com/v2/repositories/ptrxyz/chemotion/tags?page_size=1024')
        return dict(
            [(x['name'][:-(len(version) + 1)], f"ptrxyz/chemotion:{x['name']}") for x in res.json()['results'] if
             x['name'].endswith(version)])

    @staticmethod
    def insert_dataset(ols_term_id: str, pguser='postgres', port=5432, host='localhost', pgpassword='postgres',
                       database='chemotion'):

        ols_term_id_num = ols_term_id.split(':')[-1]
        ols_link = f'https://www.ebi.ac.uk/ols4/api/v2/ontologies/chmo/classes/http%253A%252F%252Fpurl.obolibrary.org%252Fobo%252FCHMO_{ols_term_id_num}?lang=en&includeObsoleteEntities=t'
        res = requests.get(ols_link)
        if res.status_code != 200:
            raise ConnectionError(f'Unable to find: {ols_term_id} ({ols_link})')
        res_json = res.json()
        ols_term_id = res_json['curie']
        label = res_json['label']
        definition = res_json['definition']['value']
        created_at = datetime.now().strftime("%Y-%m-%d, %H:%M:%S.%f")
        sql_command = f"INSERT INTO public.dataset_klasses (ols_term_id, label, \"desc\", created_by, created_at) VALUES ('{ols_term_id}', '{label}', '{definition}', 1, '{created_at}')";
        my_env = os.environ.copy()
        my_env["PGPASSWORD"] = pgpassword
        command = ['psql', '-U', pguser, '-h', host, '-p', f'{port}', '-d', database, '-c', sql_command]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, env=my_env)
        proc.communicate()



import copy
import re
import string
import uuid
from enum import Enum

from chemotion_api.labImotion.items.options import ElementType
from chemotion_api.labImotion.items.layer import Layer
from chemotion_api.labImotion.items.validation import validate_selection_options, validate_generic_properties, \
    validate_generic_dataset_properties, validate_generic_segment_properties
from chemotion_api.labImotion.structure import StructureRegistry
from chemotion_api.labImotion.utils import ChemotionSerializable
from chemotion_api.utils import FixedDict


class GenericProperties(ChemotionSerializable):
    def __init__(self, json_data: dict, element_type: ElementType, sr: StructureRegistry):
        self._json_data = json_data
        self._element_type = element_type
        self._pkg = FixedDict(copy.deepcopy(json_data.get('pkg', {})))
        self._select_options = {}
        self._identifier = None
        self._flow_object = []
        self._version = None
        self._uuid = ''
        self._klass = ElementType.ELEMENT
        self._layers = []
        self._sr = sr

        for key, value in json_data.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def to_dict(self) -> dict:
        for key, _ in self._json_data.items():
            if hasattr(self, key) and key != 'pkg':
                value = getattr(self, key)
                if isinstance(value, Enum):
                    value = value.value
                if key == 'layers':
                    self._json_data[key] = {x.key: x.to_dict() for x in value}
                else:
                    self._json_data[key] = value
        return self._json_data

    def validate(self, json_data: dict = None):
        if json_data is None:
            json_data = self._json_data
        if self._element_type == ElementType.DATASET:
            validate_generic_dataset_properties(json_data)
        if self._element_type == ElementType.SEGMENT:
            validate_generic_segment_properties(json_data)
        else:
            validate_generic_properties(json_data)

    def get_layer_by_label(self, label: str) -> Layer:
        return next(l for l in self.layers if l.label == label)

    def get_layer_by_key(self, key: str) -> Layer:
        return next(l for l in self.layers if l.key == key)

    @property
    def layers(self) -> list[Layer]:
        return self._layers

    @layers.setter
    def layers(self, values: dict | list[Layer]):
        if isinstance(values, dict):
            self._layers = []
            for key, value in values.items():
                self._layers.append(Layer(self._sr, value))
        else:
            if not isinstance(values, list):
                raise TypeError(f"layers {type(values)} not supported")
            for value in values:
                if not isinstance(value, Layer):
                    raise TypeError(f"layers -> list[{type(value)}] not supported")
                self._layers.append(value)

        self._layers.sort(key=lambda x: x.position)

    @property
    def version(self) -> str | None:
        return self._version

    @version.setter
    def version(self, value: str | None):
        if value is None:
            return
        if not isinstance(value, str):
            raise TypeError(f"version {type(value)} not supported")
        if self._version is not None:
            raise ValueError(f"version {self._version} already set")
        self._version = value

    @property
    def identifier(self) -> str:
        return self._identifier

    @identifier.setter
    def identifier(self, value: str | None):
        if value is None:
            self._identifier = uuid.uuid4().__str__()
            return
        if not isinstance(value, str):
            raise TypeError(f"identifier {type(value)} not supported")
        if self._identifier is not None:
            raise ValueError(f"identifier {self._identifier} already set")
        self._identifier = value

    @property
    def uuid(self) -> str:
        return self._uuid

    @uuid.setter
    def uuid(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"uuid {type(value)} not supported")
        if self._uuid != '':
            raise ValueError(f"uuid {self._uuid} already set")
        self._uuid = value

    @property
    def package_info(self):
        return self._pkg

    @property
    def klass(self) -> ElementType:
        return self._klass

    @klass.setter
    def klass(self, value: ElementType | str):
        if isinstance(value, str):
            value = ElementType(value)
        if not isinstance(value, ElementType):
            raise TypeError(f"klass {type(value)} not supported")
        self._klass = value

    @property
    def select_options(self) -> dict[str, dict[str, list[dict[str, str]]]]:
        return self._select_options

    @select_options.setter
    def select_options(self, value: dict[str, dict[str, list[dict[str, str]]]]):
        if not isinstance(value, dict):
            raise TypeError(f"select_options {type(value)} not supported")
        validate_selection_options(value)
        self._select_options = value

    def add_select_option(self, key: str, vals: list[str]):
        if re.search(r'^[a-z][a-z_]*[a-z]$', key) is None:
            raise ValueError(f"Option key: 1) Must start with a lowercase letter 2) Can have lowercase letters or underscores in the middle 3) Must end with a lowercase letter 4) No special characters ($ ! % …) → only lowercase and _ allowed")
        self.select_options[key] = {'options': [{'key': val, 'label': val} for val in vals]}

    def add_new_layer(self, label: str, key: str = '', idx: int = -1) -> Layer:
        all_keys = [l.key for l in self.layers]
        if key == '':
            key = re.sub(r'[^a-z]', '_', label.lower())
        t_key = key
        i = 0
        while key in all_keys or key.endswith('_'):
            key = t_key + string.ascii_lowercase[i]
            i += 1
        new_layer = Layer.generate_new_layer(self._sr, key, label)
        if idx < 0 :
            self.layers.append(new_layer)
        else:
            self.layers.insert(idx, new_layer)
        self.update_layer_position()

        return new_layer

    def update_layer_position(self):
        i = 10
        for layer in self.layers:
            if not layer.wf:
                layer.position = i
                i += 10




from enum import Enum


class EnumConnector(Enum):
    AND = 'AND'
    OR = 'OR'


class EnumMatch(Enum):
    LIKE = 'ILIKE'
    NOT_LIKE = 'NOT ILIKE'
    LESS_THAN = '<'
    BIGGER_THAN = '>'
    EXACT = '='

def new_base_field(match: EnumMatch, connector: str, value: str):
    return {
        "link": connector,
        "match": match.value,
        "value": value,
        "smiles": "",
        "sub_values": [],
        "unit": "",
        "validationState": None
    }


from chemotion_api.connection import Connection


class AdminMenu:
    def __init__(self, session: Connection):
        self._session = session

    def greate_device(self, first_name: str, last_name: str, name_abbreviation: str):
        payload = {'first_name': first_name,
                   'last_name': last_name,
                   'rootType': "Device",
                   "name_abbreviation": name_abbreviation}

        res = self._session.post('/api/v1/admin/group_device/create',
                                 headers=self._session.get_default_session_header(),
                                 data=payload)

        if res.status_code == 401:
            raise PermissionError('Not allowed to greate device (only Admin)')
        elif res.status_code != 201 and res.status_code != 200:
            raise ConnectionError(f"Error ({res.status_code}) \n{res.text}")

        return res.json()


import json
from abc import ABC, abstractmethod
from typing import Self

from chemotion_api.element_manager import ElementManager
from chemotion_api.elements import ElementSet
from chemotion_api.connection import Connection
from chemotion_api.search.utils import EnumMatch, EnumConnector


class SearchHelperBase(ABC):
    element_type_res: str
    element_type: str

    def __init__(self, connection: Connection, element_manager: ElementManager, root_collection_id: int):
        self._connection = connection
        self._element_manager = element_manager
        self._collection_id = root_collection_id
        self._per_page = 15
        self._payload = {
            "selection": {
                "elementType": "advanced",
                "advanced_params": [

                ],
                "search_by_method": "advanced",
                "page_size": self._per_page
            },
            "collection_id": self._collection_id,
            "page": 1,
            "per_page": self._per_page,
            "is_sync": False,
            "molecule_sort": True,
            "is_public": False
        }

    def in_collection(self, collection_id: int) -> Self:
        self._collection_id = collection_id
        return self

    def find_max(self, per_page: int) -> Self:
        self._per_page = per_page
        return self

    def add_search_condition(self, match: EnumMatch, connector: EnumConnector, **field_values) -> Self:
        for field_name, field_value in field_values.items():
            if field_value is not None:
                _connector = connector.value if len(
                    self._payload["selection"]["advanced_params"]) > 0 else ''
                self._payload["selection"]["advanced_params"].append(
                    self.new_field(match, _connector, field_name, field_value))
        return self

    def _send_request(self) -> list[dict]:
        self._payload['selection']['page_size'] = self._per_page
        self._payload['per_page'] = self._per_page
        self._payload['collection_id'] = self._collection_id
        with open('test.json', 'w') as file:
            file.write(json.dumps(self._payload))
        res = self._connection.post('/api/v1/search/advanced', json=self._payload)
        if res.status_code != 201:
            raise ConnectionError(res.status_code)
        return res.json()[self.element_type_res]['elements']

    def request(self) -> ElementSet:
        e = ElementSet(self._connection, self._element_manager.all_classes.get(self.element_type), self._collection_id,
                       False, True)

        for val in self._send_request():
            e.new_element(val)

        return e

    @abstractmethod
    def new_field(self, match: EnumMatch, connector: str,
                  field_name: str, value: any) -> dict[str, any]:
        ...


import abc
import json
import re
from chemotion_api.connection import Connection


class User(abc.ABC):
    @classmethod
    def load_me(cls, session: Connection):
        user_url = '/api/v1/users/current.json'
        res = session.get(user_url)
        if res.status_code == 401:
            raise PermissionError('Not allowed to fetch user (Login first)')
        elif res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        data = res.json()['user']
        return cls.factor_user(session, data)

    @classmethod
    def factor_user(cls, session: Connection, data: dict):
        if cls._is_device(data['type']):
            user = Device(session)
        elif cls._is_admin(data['type']):
            user = Admin(session)
        elif cls._is_group(data['type']):
            user = Group(session)
        else:
            user = Person(session)

        user.populate(data)

        return user

    def __init__(self, session: Connection):
        self._session = session
        self.devices = []
        self.groups = []
        self.id: int | None = None
        self.email: str | None = None
        self.user_type: str | None = None
        self.first_name: str | None = None
        self.last_name: str | None = None
        self.initials: str | None = None

    @property
    def name_abbreviation(self):
        return self.initials

    @name_abbreviation.setter
    def name_abbreviation(self, initials):
        self.initials = initials

    @property
    def name(self):
        if self.last_name is None:
            return self.first_name
        return f"{self.first_name} {self.last_name}"

    @name.setter
    def name(self, name):
        self.first_name = name

    def populate(self, json_contnet):
        self.user_type = json_contnet.get('type')
        for (key, val) in json_contnet.items():
            if hasattr(self, key):
                if re.match(r'^(\"[^\"]+\"=>\"[^\"]+\",? ?)*$', str(val)) is not None:
                    val = json.loads(f"{{ {val.replace('=>', ':')} }}")
                setattr(self, key, val)
        return self

    def is_admin(self):
        return self._is_admin(self.user_type)

    def is_device(self):
        return self._is_device(self.user_type)

    def is_group(self):
        return self._is_group(self.user_type)

    def is_person(self):
        return self._is_person(self.user_type)

    @classmethod
    def _is_admin(cls, type):
        return type.lower() == 'admin'

    @classmethod
    def _is_device(cls, type):
        return type.lower() == 'device'

    @classmethod
    def _is_group(cls, type):
        return type.lower() == 'group'

    @classmethod
    def _is_person(cls, type):
        return type.lower() == 'person'

    @classmethod
    def find_user_by_name(cls, _session: Connection, name: str) -> list:
        res = _session.get(f'/api/v1/users/name.json?name={name}&type=Person')
        if res.status_code != 200:
            raise ConnectionError(f"{res.status_code} -> {res.text}")
        return [Person(_session).populate(x) for x in res.json()["users"]]


class Group(User):

    def __init__(self, session: Connection):
        super().__init__(session)

    def add_users(self, *users: User):
        return self.add_users_by_id(*[u.id for u in users if u.is_person()])

    def populate(self, json_contnet):
        json_contnet['type'] = json_contnet.get('type', 'group')
        return super().populate(json_contnet)

    def add_users_by_id(self, *users: int):
        payload = {
            'destroy_group': False,
            'add_users': users,
            'id': self.id
        }
        res = self._session.put(f'/api/v1/groups/upd/{self.id}', data=payload)
        if res.status_code != 200:
            raise ConnectionError(f'{res.status_code} -> {res.text}')



class Person(User):

    def __init__(self, session: Connection):
        super().__init__(session)
        self.samples_count: int | None = None
        self.reactions_count: int | None = None
        self.reaction_name_prefix: str | None = None
        self.layout: dict[str:str] = None
        self.unconfirmed_email: str | None = None
        self.confirmed_at: str | None = None
        self.current_sign_in_at: str | None = None
        self.locked_at = None
        self.is_templates_moderator: bool | None = None
        self.molecule_editor: bool | None = None
        self.account_active: bool | None = None
        self.matrix: int | None = None
        self.counters: dict[str:str] | None = None
        self.generic_admin: dict[str:bool] = None

    def populate(self, json_contnet):
        json_contnet['type'] = json_contnet.get('type', 'person')
        return super().populate(json_contnet)

    def get_next_short_label(self, type_name):
        if type_name == 'sample':
            return '{}-{}'.format(self.initials, self.counters[type_name + 's'])
        if type_name == 'reaction':
            return '{}-R{}'.format(self.initials, self.counters[type_name + 's'])
        return f"{self.initials}-{type_name}{self.counters.get(type_name + 's', self.counters.get(type_name, 0))}"

    def create_group(self, first_name: str, last_name: str, name_abbreviation: str) -> Group:
        data = {
            'group_param': {
                "first_name": first_name,
                "last_name": last_name,
                "name_abbreviation": name_abbreviation,
                "users": [
                    self.id
                ]
            }
        }
        url = "/api/v1/groups/create"
        res = self._session.post(url, data=data)
        if res.status_code != 201:
            raise ConnectionError(f'{res.status_code} -> {res.text}')
        data = res.json()
        if 'error' in data:
            raise ConnectionError(f'{res.text}')
        data['type'] = 'group'
        return Group(self._session).populate(data['group'])


class Device(User):

    def __init__(self, session: Connection):
        super().__init__(session)
        self.is_super_device: bool | None = None

    def get_jwt(self):
        return DeviceManager(self._session).get_jwt_for_device(self.id)

    def delete(self):
        url = f"/api/v1/admin/group_device/update/{self.id}"
        res = self._session.put(url, data={
            "action": "RootDel",
            "rootType": "Device",
            "id": self.id,
            "destroy_obj": True,
            "rm_users": []
        })
        if res.status_code == 401:
            raise PermissionError('Not allowed to delete device (Only for super devices or admins)')
        elif res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))


class Admin(Person):

    def __init__(self, session: Connection):
        super().__init__(session)
        self.users: list[User] = []
        self.devices: list[Device] = []
        self.groups: list[Group] = []
        self.admins: list[Admin] = []

    def populate(self, json_contnet):
        json_contnet['type'] = json_contnet.get('type', 'admin')
        return super().populate(json_contnet)

    def fetchUsers(self):
        user_url = f'/api/v1/admin_user/listUsers/all.json'
        res = self._session.get(user_url)
        if res.status_code == 401:
            raise PermissionError('Not allowed to fetch user (Login first)')
        elif res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        self.users: list[User] = []
        self.devices: list[Device] = []
        self.groups: list[Group] = []
        self.admins: list[Admin] = []
        for u_data in res.json()['users']:
            user = self.factor_user(self._session, u_data)
            if user.is_admin():
                self.admins.append(user)
            elif user.is_device():
                self.devices.append(user)
            elif user.is_group():
                self.groups.append(user)
            else:
                self.users.append(user)

    def create_device(self, first_name: str, last_name: str, name_abbreviation: str) -> Device:
        data = self._create_group_device(first_name=first_name, last_name=last_name,
                                         name_abbreviation=name_abbreviation, rootType='Device')
        return Device(self._session).populate(data)

    def create_group(self, first_name: str, last_name: str, name_abbreviation: str) -> Group:
        data = self._create_group_device(first_name=first_name, last_name=last_name,
                                         name_abbreviation=name_abbreviation, rootType='Group')
        return Group(self._session).populate(data)

    def create_person(self, first_name: str, last_name: str, name_abbreviation: str, password: str,
                      email: str) -> Person:
        data = {
            "email": email,
            "first_name": first_name,
            "last_name": last_name,
            "name_abbreviation": name_abbreviation,
            "password": password,
            "type": "Person"
        }
        url = "/api/v1/admin/users/"
        res = self._session.post(url, data=data)
        if res.status_code != 201:
            raise ConnectionError(f'{res.status_code} -> {res.text}')

        return [x for x in self.find_user_by_name(self._session, f"{first_name} {last_name}") if
                x.name_abbreviation == name_abbreviation][0]

    def _create_group_device(self, first_name: str, last_name: str, name_abbreviation: str, rootType: str):
        url = '/api/v1/admin/group_device/create'
        group = dict(first_name=first_name, last_name=last_name, name_abbreviation=name_abbreviation, rootType=rootType)
        res = self._session.post(url, data=group)
        if res.status_code != 201:
            raise ConnectionError(f'{res.status_code} -> {res.text}')
        data = res.json()
        if 'error' in data:
            raise ConnectionError(f'{res.text}')
        return data

    def fetchDevices(self) -> list[Device]:
        u_type = 'Device'
        d_list = self.fetchGroupAndDevice(u_type)
        self.devices = [Device(self._session).populate(x | {'type': u_type}) for x in d_list]
        return self.devices

    def fetchGroups(self) -> list[Group]:
        u_type = 'Group'
        g_list = self.fetchGroupAndDevice(u_type)
        self.groups = [Group(self._session).populate(x | {'type': u_type}) for x in g_list]
        return self.groups

    def fetchGroupAndDevice(self, type: str) -> list[dict]:
        user_url = f'/api/v1/admin/group_device/list?type={type}'
        res = self._session.get(user_url)
        if res.status_code == 401:
            raise PermissionError('Not allowed to fetch user (Login first)')
        elif res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        return res.json()['list']


class DeviceManager:
    def __init__(self, session: Connection):
        self._session = session

    def get_jwt_for_device(self, id: int):
        user_url = f'/api/v1/devices/remote/jwt/{id}'
        res = self._session.get(user_url)
        if res.status_code == 401:
            raise PermissionError('Not allowed to fetch JWT (Only for super devices or admins)')
        elif res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        return res.json()['token']

    def create_new_device(self, first_name: str, last_name: str, name_abbreviation: str, email: str = None) -> Device:
        user_url = f'/api/v1/devices/remote/create'
        data = {
            'first_name': first_name,
            'last_name': last_name,
            'name_abbreviation': name_abbreviation,
        }
        if email is not None: data['email'] = email
        res = self._session.post(user_url, data=data)
        if res.status_code == 401:
            raise PermissionError('Create device not is allowed! (Only for super devices or admins)')
        elif res.status_code != 201 or res.json().get('error') is not None:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        return Device(self._session).populate(res.json())

    def add_users_to_group(self, group: Group, *users: User):
        return self.add_users_to_group_by_id(group, *[u.id for u in users if u.is_person()])

    def add_users_to_group_by_id(self, group: Group, *users: int):
        payload = {
            'action': "NodeAdd",
            'actionType': "Person",
            'add_users': users,
            'id': group.id,
            'rootType': "Group",
        }
        res = self._session.put(f'/api/v1/admin/group_device/update/{group.id}', data=payload)
        if res.status_code != 200:
            raise ConnectionError(f'{res.status_code} -> {res.text}')



from typing import Literal, Optional

from chemotion_api.elements import ElementSet
from chemotion_api.search.search_helper import SearchHelperBase
from chemotion_api.search.utils import EnumMatch, EnumConnector, new_base_field


def field_details(field_name: Literal['name', 'short_label', 'external_label', 'inventory_label', 'xref', 'content']):
    if field_name == 'name':
        filed_type = {
            "column": "name",
            "label": "Sample name",
            "type": "text",
            "advanced": True
        }
    elif field_name == 'short_label':
        filed_type = {
            "column": "short_label",
            "label": "Short Label",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'external_label':
        filed_type = {
            "column": "external_label",
            "label": "External Label",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'inventory_label':
        filed_type = {
            "column": "xref",
            "opt": "inventory_label",
            "label": "Inventory Label",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'xref':
        filed_type = {
            "column": "xref",
            "opt": "cas",
            "label": "CAS",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'content':
        filed_type = {
            "column": "content",
            "label": "Private Note",
            "type": "text",
            "advanced": True
        }

    else:
        raise ValueError(f'Field "{field_name}" not supported.')

    return {
        'field': filed_type,
        'unit': ''
    }


class SampleSearcher(SearchHelperBase):
    element_type_res = "samples"
    element_type = "sample"

    def add_search_condition(self, match: EnumMatch, connector: EnumConnector, name: Optional[str] = None,
                             short_label: Optional[str] = None, external_label: Optional[str] = None,
                             inventory_label: Optional[str] = None, cas: Optional[str] = None,
                             private_note: Optional[str] = None):
        field_values = {"name": name, "short_label": short_label, "external_label": external_label,
                        "inventory_label": inventory_label, "xref": cas, "content": private_note}
        return super().add_search_condition(match, connector, **field_values)

    def new_field(self, match: EnumMatch, connector: str,
                  field_name: Literal['name', 'short_label', 'external_label', 'inventory_label', 'xref', 'content'],
                  value: str) -> dict:
        return new_base_field(match, connector, value) | {
            "table": "samples",
            "element_id": 0,
        } | field_details(field_name)




import copy

from chemotion_api.labImotion.items.validation.registry import SchemaRegistry
from chemotion_api.labImotion.items.validation.schemas.schema_generc_element import generic_element_schema
from chemotion_api.labImotion.items.validation.schemas.schema_properties import properties_schema as properties

generic_schema = copy.deepcopy(generic_element_schema)
generic_schema['$id'] = 'chemotion://generic/dataset/draft-01'
generic_schema['title'] = "Schema for generic dataset"
generic_schema['properties']['properties_template']['$ref'] = 'chemotion://generic/dataset_properties/draft-01'
generic_schema['properties']['properties_release']['$ref'] = 'chemotion://generic/dataset_properties/draft-01'


generic_schema['properties'] |= {
    "version": {
        "type": ["string", 'null']
    },
    "released_at": {
        "type": ["string", 'null']
    },
    "updated_by": {
        "type": ["number", 'null']
    },
    "ols_term_id": {
        "type": "string"
    }
}

generic_schema['required'] = [x for x in generic_schema['required'] if x not in ['pkg']]
generic_schema['required'].append('ols_term_id')

generic_props_schema = copy.deepcopy(properties)
generic_props_schema['$id'] = 'chemotion://generic/dataset_properties/draft-01'
generic_props_schema['title'] = "Schema for generic dataset properties"
generic_props_schema['required'] = [x for x in generic_props_schema['required'] if x not in ['pkg', 'version', 'identifier', 'uuid', 'klass', 'layers', 'select_options']]

SchemaRegistry.instance().register(generic_schema)
SchemaRegistry.instance().register(generic_props_schema)



from datetime import datetime
from typing import Optional

from chemotion_api.elements.abstract_element import AbstractElement
from chemotion_api.elements.generic_element import GenericElement
from chemotion_api.elements.sample import Sample
from chemotion_api.elements.reaction import Reaction
from chemotion_api.elements.wellplate import Wellplate
from chemotion_api.elements.research_plan import ResearchPlan
from chemotion_api.generic_segments import GenericSegments
from chemotion_api.connection import Connection
from requests.exceptions import RequestException


class ElementSet(list):
    """
    The ElementSet class allows you to access all elements of a type in a collection.
    You can iterate over them or filter them using an IRI filter.
    """

    def __init__(self, session: Connection, element_type: dict,
                 collection_id: int = None, collection_is_sync: bool = False, fixed_length:bool = False):
        super().__init__()
        self._load_options = {}
        self._session = session
        self._element_type = element_type
        self._collection_id = collection_id
        self._collection_is_sync = collection_is_sync
        self._page = 1
        self.max_page = 1
        self._per_page = 10
        self._iri_filter = None
        self._fixed_length = fixed_length

    def set_iri_filter(self, iri):
        self._iri_filter = iri

    def __len__(self):
        if self._iri_filter is None or self._fixed_length:
            return super().__len__()
        else:
            raise TypeError("Length cannot be predicted if you use IRI typs")

    def __getitem__(self, item):
        if isinstance(item, slice):
            start = item.start if item.start is not None else 0
            stop = item.stop if item.stop is not None else len(self)
            last = stop - start
            if last <= 0 or stop < 0 or start < 0:
                raise TypeError(f"Negative index is not supported")
            per_page = self._per_page
            if item.start is not None:
                self._per_page = start
                start_page = 2
            else:
                self._per_page = min(10, last)
                start_page = 1
            res_list = []
            for x in self.iter_elements(start_page):
                res_list.append(x)
                if (len(res_list) == last):
                    break
            self._per_page = per_page
            return res_list

        if self._iri_filter is not None:
            raise TypeError(f"Index is prohibited if you use IRI types")
        if item < 0:
            raise TypeError(f"Negative index is not supported")
        page = (item // self._per_page) + 1
        if page != self._page:
            self._set_page(page)
        idx = item % self._per_page
        elm: AbstractElement = super().__getitem__(idx)
        elm.load()
        return elm

    def __iter__(self):
        for x in self.iter_elements():
            yield x

    @property
    def page(self):
        return self._page

    @page.setter
    def page(self, page: int):
        self._set_page(page)

    def _set_page(self, page: int) -> bool:
        if self._fixed_length:
            self._page = page
            return page == 1
        if page > 0 and page <= self.max_page:
            self._page = page
            self.load_elements(**self._load_options)
            return True
        return False

    def iter_elements(self, start_page=1):
        self._page = start_page - 1
        while self._set_page(self.page + 1):
            for elem in super().__iter__():
                if self._iri_filter is None or self._iri_filter == elem.json_ld['@type']:
                    yield elem

    def iter_pages(self):
        self._page = 0
        while self._set_page(self.page + 1):
            yield self

    def next_page(self):
        self._set_page(self.page + 1)
        return self

    def prev_page(self):
        self._set_page(self.page - 1)
        return self

    def load_elements(self, per_page: Optional[int] = None, updated_to_date: Optional[datetime] = None,
                      updated_from_date: Optional[datetime] = None, product_only: bool = False):
        if per_page is not None:
            self._per_page = per_page
        if self._collection_id is None:
            raise ValueError('load_elements only works if collection_id is set!')

        self._load_options = {
            'updated_to_date': updated_to_date,
            'updated_from_date': updated_from_date,
            'product_only': product_only,
            'per_page': per_page,
        }

        segments = GenericSegments(self._session)
        payload = {'page': self.page,
                   'per_page': self._per_page,
                   'filter_created_at': False,
                   'product_only': product_only,
                   'el_type': self._element_type['name']}

        if updated_from_date:
            payload['from_date'] = int(updated_from_date.timestamp())

        if updated_to_date:
            payload['to_date'] = int(updated_to_date.timestamp())

        if self._collection_is_sync:
            payload['sync_collection_id'] = self._collection_id
        else:
            payload['collection_id'] = self._collection_id
        res = self._session.get(self._get_url() + '.json',
                                data=payload)
        if res.status_code != 200:
            raise RequestException('{} -> {}'.format(res.status_code, res.text))

        try:
            self.max_page = int(res.headers.get('X-Total-Pages', 1))
            self._length = int(res.headers.get('X-Total', 0))
        except:
            pass
        self.clear()
        elements = res.json().get(self._get_result_key() + 's', [])
        for json_data in elements:
            self.append(self._get_element_class()(segments, self._session, json_data=json_data))

    def load_element(self, id: int) -> AbstractElement:
        segments = GenericSegments(self._session)
        s = self._get_element_class()(segments, self._session, id=id,
                                      element_type=self._element_type['name'])
        self.append(s)
        return s

    def new_element(self, json_data: dict, **kwargs) -> AbstractElement:
        segments = GenericSegments(self._session)
        s = self._get_element_class()(segments, self._session, json_data=json_data, **kwargs)
        self.append(s)
        return s

    def _get_element_class(self) -> AbstractElement.__class__:
        if self._element_type['is_generic']:
            return GenericElement
        elif self._element_type['name'] == 'sample':
            return Sample
        elif self._element_type['name'] == 'reaction':
            return Reaction
        elif self._element_type['name'] == 'wellplate':
            return Wellplate
        elif self._element_type['name'] == 'research_plan':
            return ResearchPlan

        raise TypeError('Generic type "{}" cannot be found'.format(self._element_type['name']))

    def _get_result_key(self):
        if self._element_type['is_generic']:
            return 'generic_element'
        return AbstractElement.get_response_key(self._element_type['name'])

    def _get_url(self):
        return AbstractElement.get_url(self._element_type['name'])



import json
from typing import Self, Callable

from chemotion_api.connection import Connection
from chemotion_api.labImotion.items.genericproperties import GenericProperties
from chemotion_api.labImotion.items.options import SaveActionType
from chemotion_api.labImotion.structure import StructureRegistry
from chemotion_api.labImotion.utils import ChemotionSerializable, generate_pkg_info


class GenericObj(ChemotionSerializable):
    url = None
    url_create = None
    url_upload = None
    url_activate = None
    url_save = None
    url_save_template = None
    element_type = None

    def __init__(self, json_data: dict, connection: Connection):
        self._json_data = json_data
        self._connection = connection
        self._id = json_data['id']
        self._version = json_data['version']
        self._identifier = json_data['identifier']
        self._uuid = json_data['uuid']
        self._is_active = json_data['is_active']
        self._label = json_data['label']
        self._desc = json_data['desc']
        self._is_generic = json_data.get('is_generic', True)
        sr = StructureRegistry(self)
        self._properties = None
        if self.is_generic:
            self._properties = GenericProperties(json_data['properties_template'], self.element_type, sr)
            self._release = GenericProperties(json_data['properties_release'], self.element_type, sr)

    def to_dict(self):
        if self.is_generic:
            self._json_data |= {
                'uuid': self._identifier,
                'label': self.label,
                'desc': self.desc,
                'properties_template': self._properties.to_dict()
            }
        return self._json_data

    def save(self, save_type_action: SaveActionType = SaveActionType.DRAFT):
        if self.url_save_template is not None:
            self.save_template(save_type_action)
        if self.url_save is not None:
            self.save_generale_properties()

    def save_template(self, save_type_action: SaveActionType = SaveActionType.DRAFT):
        if self.url_save_template is None:
            raise TypeError('Save template is not supported')
        res = self._connection.post(self.url_save_template, data=self.to_dict() | {
            "release": save_type_action.value,
            "klass": self.element_type.value,
            "version": None
        })
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")

    def save_generale_properties(self):
        if self.url_save is None:
            raise TypeError('Save template is not supported')
        res = self._connection.post(self.url_save, data=self.to_dict())
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")

    def activate(self):
        self.is_active = True

    def deactivate(self):
        self.is_active = False

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"label {type(value)} not supported")
        self._label = value
        
    @property
    def desc(self) -> str:
        return self._desc
    
    @desc.setter
    def desc(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"desc {type(value)} not supported")
        self._desc = value

    @property
    def version(self) -> str:
        return self._version

    @property
    def is_active(self) -> bool:
        return self._is_active

    @is_active.setter
    def is_active(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"is_active {type(value)} not supported")
        res = self._connection.post(self.url_activate, data={
            'klass': self.element_type.value,
            'id': self.id,
            'is_active': value
        })
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")
        self._is_active = value

    @property
    def identifier(self) -> str:
        return self._identifier

    @identifier.setter
    def identifier(self, value: str):
        if value is None:
            return
        if not isinstance(value, str):
            raise TypeError(f"identifier {type(value)} not supported")
        self._identifier = value

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def properties(self) -> GenericProperties | None:
        return self._properties

    @property
    def release(self) -> GenericProperties | None:
        return self._release

    @property
    def id(self) -> int:
        return self._id

    @property
    def is_generic(self) -> bool:
        return self._is_generic

    def validate(self, json_data: dict = None):
        if json_data is None:
            json_data = self._json_data

        self.validator()(json_data)

    @classmethod
    def validator(cls) -> Callable[[dict], None]:
        raise NotImplementedError()

    @classmethod
    def load_all(cls, con: Connection) -> list[Self]:
        if cls.url is None:
            return []
        res = con.get(cls.url)
        if res.status_code != 200:
            raise ConnectionError(f"{res.status_code} -> {res.text}")
        return_val = list()
        klass_list = res.json()['klass']
        for obj in klass_list:
            return_val.append(cls(obj, con))
        return return_val

    @classmethod
    def upload(cls, con: Connection, file_path: str) -> list[Self]:
        if cls.url_create is None:
            return []
        with open(file_path, 'r') as f:
            json_data = json.loads(f.read())
        res = con.post(cls.url_upload, data=json_data)
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")
        return cls.load_all(con)

    @classmethod
    def create_new(cls, con: Connection, **kwargs) -> list[Self]:
        if cls.url_create is None:
            return []
        res = con.post(cls.url_create, data={
            **kwargs
        } | generate_pkg_info())
        if res.status_code != 201:
            raise ConnectionError(f"{res.status_code} -> {res.text}")

        return cls.load_all(con)


from chemotion_api.labImotion.items.fields.abstract_fields import SystemDefined, Select
from chemotion_api.labImotion.items.fields.sub_fields import SubBaseField
from chemotion_api.labImotion.items.options import SubTableFieldType
from chemotion_api.labImotion.structure import StructureRegistry

class TableSubBaseField(SubBaseField):
    _default_required_fields = ["id", "col_name", "field_type", "value"]
    required_fields = []

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._col_name = json_data.get("col_name", "")

    @property
    def col_name(self) -> str:
        return self._col_name
    
    @col_name.setter
    def col_name(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"col_name {type(value)} not supported")
        self._col_name = value


class TSDragMolecule(TableSubBaseField):
    field_type = SubTableFieldType.DRAG_MOLECULE

    @property
    def value(self) -> str:
        if self._value is None or self._value == '':
            self._value = ';'
        return self._value

    @value.setter
    def value(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"value {type(value)} not supported")
        text_display = []
        value_split = value.split(';')
        if 'inchikey' in value_split:
            text_display.append("inchikey;")
        if 'smiles' in value_split:
            text_display.append("smiles;")
        if 'iupac' in value_split:
            text_display.append("iupac;")
        if 'molecular_weight' in value_split:
            text_display.append("molecular_weight;")
        self._value = ''.join(text_display)

    def set_sample_display(self, inchikey: bool, smiles: bool, iupac: bool, molecular_weight: bool):
        text_display = []
        if inchikey:
            text_display.append("inchikey;")
        if smiles:
            text_display.append("smiles;")
        if iupac:
            text_display.append("iupac;")
        if molecular_weight:
            text_display.append("molecular_weight;")
        self.value = ''.join(text_display)


class TSDragSample(TableSubBaseField):
    field_type = SubTableFieldType.DRAG_SAMPLE

    @property
    def value(self) -> str:
        if self._value is None or self._value == '':
            self._value = ';'
        return self._value

    @value.setter
    def value(self, value: str):
        if not isinstance(value, str):
            raise TypeError(f"value {type(value)} not supported")
        text_display = []
        value_split = value.split(';')
        if 'name' in value_split:
            text_display.append("name;")
        if 'external_label' in value_split:
            text_display.append("external_label;")
        if 'molecular_weight' in value_split:
            text_display.append("molecular_weight;")
        self._value = ''.join(value)

    def set_sample_display(self, name: bool, external_label: bool, molecular_weight: bool):
        text_display = []
        if name:
            text_display.append("name;")
        if external_label:
            text_display.append("external_label;")
        if molecular_weight:
            text_display.append("molecular_weight;")
        self.value = ''.join(text_display)



class TSSelect(TableSubBaseField, Select):
    field_type = SubTableFieldType.SELECT
    required_fields = ['option_layers']

    def __init__(self, sr: StructureRegistry,  layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        Select.__init__(self, sr, json_data)


class TSSystemDefined(TableSubBaseField, SystemDefined):
    field_type = SubTableFieldType.SYSTEM_DEFINED
    required_fields = ['option_layers', 'value_system']

    def __init__(self, sr: StructureRegistry,  layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        SystemDefined.__init__(self, json_data)


class TSText(TableSubBaseField):
    field_type = SubTableFieldType.TEXT



from enum import Enum
from typing import Optional


class GasType(Enum):
    OFF = 'off'
    GAS = 'gas'
    CATALYST = 'catalyst'
    FEEDSTOCK = 'feedstock'


class QuantityUnit(Enum):
    Liter = 'l'
    Mole = 'mol'
    Gram = 'g'
    Equivalent = 'Equivalent'
    Yield = 'Yield'

    @classmethod
    def from_string(cls, input_string: str):
        for unit in list(cls):
            if unit.value.lower() == input_string.lower():
                return unit
        return None


def _get_from_gram(current_value: float, output_unit: QuantityUnit, molecular_weight: Optional[float], purity: float,
                   density: float = 0, molarity: float = 0) -> float:
    if output_unit == QuantityUnit.Gram:
        return current_value
    if output_unit == QuantityUnit.Mole:
        if molecular_weight is None:
            raise KeyError(f'Molecular weight needs to be known!')
        try:
            return current_value * purity / molecular_weight
        except ZeroDivisionError:
            return float('NaN')
    if output_unit == QuantityUnit.Liter:
        if molarity != 0:
            return _get_from_gram(current_value, QuantityUnit.Mole, molecular_weight, purity, density,
                                  molarity) / molarity
        elif density != 0:
            return current_value / (density * 1000)

    return float('NaN')


def _get_from_mol(current_value: float, output_unit: QuantityUnit, molecular_weight: Optional[float], purity: float,
                  density: float = 0, molarity: float = 0) -> float:
    if output_unit == QuantityUnit.Mole:
        return current_value
    if output_unit == QuantityUnit.Gram:
        if molecular_weight is None:
            raise KeyError(f'Molecular weight needs to be known!')
        try:
            return current_value * molecular_weight / purity
        except ZeroDivisionError:
            return float('NaN')
    if output_unit == QuantityUnit.Liter:
        if molarity != 0:
            return current_value / molarity
        elif density != 0:
            return _get_from_mol(current_value, QuantityUnit.Gram, molecular_weight, purity, density, molarity) / (
                        density * 1000)

    return float('NaN')


def _get_from_liter(current_value: float, output_unit: QuantityUnit, molecular_weight: Optional[float], purity: float,
                    density: float = 0, molarity: float = 0) -> float:
    if output_unit == QuantityUnit.Liter:
        return current_value
    if output_unit == QuantityUnit.Mole:
        if molarity != 0:
            return current_value * molarity
        elif density != 0:
            current_gram = current_value * (density * 1000)
            return _get_from_gram(current_gram, QuantityUnit.Mole, molecular_weight, purity, density, molarity)
    if output_unit == QuantityUnit.Gram:
        if molarity != 0:
            current_gram = current_value * molarity
            return _get_from_mol(current_gram, QuantityUnit.Gram, molecular_weight, 1, density, molarity)
        elif density != 0:
            return current_value * (density * 1000)

        return current_value

    return float('NaN')


def calculate_quantity(current_value: float, current_unit: QuantityUnit, output_unit: QuantityUnit,
                       molecular_weight: Optional[float], purity: float, density: float = 0, molarity: float = 0,
                       gas_type: GasType = GasType.OFF, vessel_size: float = 0, temperature_in_k: float = 0, ppm: float = 0, time_in_h: float = 0) -> float:
    if current_value is None:
        current_value = 0.0

    if gas_type != GasType.OFF:
        return GasCalculations.calculate_quantity(current_value, current_unit, output_unit, molecular_weight, purity, density,
                                           molarity, gas_type, vessel_size, temperature_in_k, ppm, time_in_h)
    if current_unit == QuantityUnit.Gram:
        return _get_from_gram(current_value, output_unit, molecular_weight, purity, density, molarity)
    if current_unit == QuantityUnit.Mole:
        return _get_from_mol(current_value, output_unit, molecular_weight, purity, density, molarity)
    if current_unit == QuantityUnit.Liter:
        return _get_from_liter(current_value, output_unit, molecular_weight, purity, density, molarity)
    return float('NaN')


class GasCalculations:
    @classmethod
    def calculate_quantity(cls, current_value: float, current_unit: QuantityUnit, output_unit: QuantityUnit,
                           molecular_weight: Optional[float], purity: float, density: float = 0, molarity: float = 0,
                           gas_type: GasType = GasType.OFF, vessel_size: float = 0, temperature_in_k: float = 0, ppm: float = 0, time_in_h: float = 0) -> float:
        if gas_type == GasType.CATALYST:
            # Catalyst is calculated as normal material
            return calculate_quantity(current_value, current_unit, output_unit, molecular_weight, purity, density,
                                           molarity, GasType.OFF)
        if gas_type == GasType.FEEDSTOCK:
            if current_unit == QuantityUnit.Mole:
                return cls._feedstock_from_mol(current_value, output_unit, purity)
            if current_unit == QuantityUnit.Liter:
                return cls._feedstock_from_liter(current_value, output_unit, purity)
        if gas_type == GasType.GAS:
            return cls._gas_converter(output_unit, molecular_weight, vessel_size, temperature_in_k, ppm, time_in_h )
        return float('NaN')


    IDEAL_GAS_CONSTANT = 0.0821
    PARTS_PER_MILLION_FACTOR = 1_000_000
    DEFAULT_TEMPERATURE_IN_KELVIN = 294

    @classmethod
    def _feedstock_from_liter(cls, current_value, output_unit, purity):
        if output_unit == QuantityUnit.Liter:
            return current_value
        if output_unit == QuantityUnit.Mole:
            return (current_value * purity) / (cls.IDEAL_GAS_CONSTANT * cls.DEFAULT_TEMPERATURE_IN_KELVIN)
        return float('NaN')

    @classmethod
    def _feedstock_from_mol(cls, current_value, output_unit, purity):
        if output_unit == QuantityUnit.Mole:
            return current_value
        if output_unit == QuantityUnit.Liter:
            return  (cls.IDEAL_GAS_CONSTANT * cls.DEFAULT_TEMPERATURE_IN_KELVIN) / (current_value * purity)
        return float('NaN')

    @classmethod
    def _gas_converter(cls, output_unit, molecular_weight, vessel_size,  temperature_in_k: float = 0, ppm: float = 0, time_in_h: float = 0):
        if vessel_size is None:
            vessel_size = 0
        if output_unit == QuantityUnit.Liter:
            return vessel_size / molecular_weight
        if output_unit == QuantityUnit.Mole:
            if ppm is None or temperature_in_k is None:
                return 0
            try:
                return (ppm * vessel_size) / (cls.IDEAL_GAS_CONSTANT * temperature_in_k * cls.PARTS_PER_MILLION_FACTOR)
            except ZeroDivisionError:
                return 0
        return float('NaN')


import hashlib
import json
import os
import uuid
from collections import OrderedDict
from collections.abc import Sequence
import magic
import requests

from chemotion_api.connection import Connection
from chemotion_api.elements.analyses_manager import AnalysesManager


class Attachment(dict):
    """
    Attachment are a dict object to read one attachment of an element.
    It allows to load Attachments as bytes or save it to a file.

    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> rp = instance.get_generic_by_label('My Generic',1)
    >>> att = rp.attachments
    >>> att.load_attachment(filename='MyAttachment.txt').save_file('.')

    :param con: Host connection object.
    :param data: attachment information (from the sever)
    """

    def __init__(self, con: Connection, data: dict[str:str]):
        self._con = con
        super().__init__(data)

    def load_file(self) -> bytes:
        """
        Loads an attachment file as binary object

        :return:  file as binary object
        """
        res = self._con.get(f"/api/v1/attachments/{self['id']}")
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))

        return res.content

    def save_file(self, directory: str = '.') -> str:
        """
        Loads an attachment file to your local filesystem

        :param directory: path to the directory where the file shell be saved

        :return: full path to the file
        """
        if not os.path.isdir(directory):
            raise NotADirectoryError(f"{directory} is not a directory")
        res = self._con.get(f"/api/v1/attachments/{self['id']}")
        if res.status_code != 200:
            raise ConnectionError('{} -> {}'.format(res.status_code, res.text))
        file_path = os.path.join(directory, self['filename'])
        with open(file_path, 'wb+') as f:
            f.write(res.content)

        return os.path.abspath(file_path)


class Attachments(Sequence):
    """
    Attachments is a container object to manage (read/write) all attachments of an element.
    Each attachemnt is a :class:`chemotion_api.elements.attachments.Attachment`.
    It allows to load Attachments or add new ones.
    Attachment can be iterated!


    Usage::

    >>> from chemotion_api import Instance
    >>> import logging
    >>> try:
    >>>     instance = Instance('http(d)://xxx.xxx.xxx').test_connection().login('<USER>', "<PASSWORD>")
    >>> except ConnectionError as e:
    >>>     logging.error(f"A connection to Chemotion ({instance.host_url}) cannot be established")
    >>> rp = instance.get_research_plan(1)
    >>> att = rp.attachments
    >>> att.load_attachment(filename='MyAttachment.txt')
    >>> att.add_file('./MySecondAttachment.txt')
    >>> rp.save()

    :param con: Host connection object.
    :param attachment_data: list of attachment information (from the sever)
    """

    def __init__(self, con: Connection, attachment_data: list[dict] = None):
        super().__init__()
        if attachment_data is None:
            attachment_data = []
        self._attachment_data = [Attachment(con, x) for x in attachment_data]
        self._new_attachment_data = []
        self._con = con
        self._to_uploads = []             # Direct inheritance

            # Extra method not required by the ABC
    def __getitem__(self, index: int) -> Attachment:
        return self.attachment_data[index]
    def __len__(self):
        return len(self.attachment_data)

    @property
    def attachment_data(self) -> list:
        """
        The Attachment data contains all information needed to save the Attachments.

        :return: List of all attachment data
        """
        return self._attachment_data + self._new_attachment_data

    def load_attachment(self, id: str | int | None = None, identifier: str | None = None,
                        filename: str | None = None) -> Attachment:
        """
        Loads an attachment from the server. One of the parameters Id, Identifier or filename must be set in order to find the desired attachment.
        If Id is set, all other parameters are ignored; if Identifier is set, filename is ignored.

        :param id: DB id of the attachment
        :param identifier: UUID of the attachment
        :param filename: name of the attachment

        :raises ValueError: If no file fits the parameters.

        :return: Attachment Object
        """
        if id is not None:
            key = 'id'
            identifier = id
        elif identifier is not None:
            key = 'identifier'
        elif filename is not None:
            key = 'filename'
            identifier = filename
        else:
            raise ValueError(f'Either id or identifier must be not None!')

        for attachment in self._attachment_data:
            if attachment[key] == identifier:
                return attachment
        raise ValueError(f'{key} {identifier} not found!')

    def add_file(self, file_path: str) -> dict:
        """
        Adds a file to the elements attachment container

        :param file_path: Path to the file to be uploaded on your device

        :return: a summary of the file
        """
        filename = os.path.basename(file_path)
        file_uuid = uuid.uuid4().__str__()
        file_magic = magic.Magic(mime=True)
        with open(file_path, 'rb') as f:
            upload_file_obj = {
                'file_uuid': file_uuid,
                'file_path': file_path,
                'filename': filename,
                'mime_type': file_magic.from_file(file_path)

            }
            chunk = f.read()

            f.close()
            data = OrderedDict(file={"preview": f"{self._con.host_url}/images/wild_card/not_available.svg"},
                               name=filename,
                               filename=filename,
                               is_deleted=False,
                               _preview="/images/wild_card/not_available.svg",
                               is_image_field=filename.lower().endswith(
                                   ('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif')),
                               filesize=len(chunk),
                               id=file_uuid,
                               is_new=True)
            data_hash_str = json.dumps(data, separators=(',', ':'))
            data_hash = hashlib.sha256(data_hash_str.encode()).hexdigest()
            data = dict(data)
            data["_checksum"] = data_hash
            data["identifier"] = file_uuid

            self._to_uploads.append(upload_file_obj)
            self._new_attachment_data.append(data)
            return data

    def save(self, id: int, element_type: str) -> requests.Response:
        """
        Updates or creates the Attachments of an Element after saving it!

        :param id: ID of the element
        :param element_type: must be in 'Element' or 'ResearchPlan'

        :return: response of the request
        """
        if element_type == 'Element':
            return self._save_generic(id)
        files = [('files[]', (x['filename'], open(x['file_path'], 'rb'), x['mime_type'])) for x in self._to_uploads]

        data = {'attfilesIdentifier[]': [x['file_uuid'] for x in self._to_uploads],
                'attachable_type': element_type,
                'attachable_id': id}
        res = self._con.post('/api/v1/attachable/update_attachments_attachable', data=data, files=files)
        return res

    def _save_generic(self, id: int):
        files = [('attfiles[]', (x['filename'], open(x['file_path'], 'rb'), x['mime_type'])) for x in self._to_uploads]

        data = {'attfilesIdentifier[]': [x['file_uuid'] for x in self._to_uploads],
                'att_type': 'Element',
                'att_id': id,
                'seInfo': {}}
        res = self._con.post('/api/v1/generic_elements/upload_generics_files', data=data, files=files)
        return res

class MutableAttachments(Attachments):
    """
    Same as :class:`chemotion_api.elements.attachments.Attachments` only add File is not allowed.
    """

    def __init__(self, con: Connection, attachment_data: list[dict] = None, *file_paths: str):
        super().__init__(con, attachment_data)
        for fp in file_paths:
            super().add_file(fp)

    def add_file(self, file_path: str) -> dict:
        raise NotImplemented("In a MutableAttachments add_file is not allowed")

    def save(self, id: int = -1, element_type: str = '') -> requests.Response | None:
        for x in self._to_uploads:
            AnalysesManager.upload_file(self._con, x['file_path'], x['file_uuid'])
        return None
